package com.dtai.gwt;

import java.net.*;
import java.io.*;

/**
 * Our dummy implementation of URLConnection used to open up
 * a TCP/IP HTTP connection when called upon with a given URL
 * of ANY protocol type. This allows the applet to easily use
 * the "getInputStream()" function, and helps handle image
 * loading.
 *
 * @version	1.1
 * @author	DTAI, Incorporated
 */
class AmHttpURLConnection extends URLConnection {

    boolean connected = false;
    InputStream instream;

    /**
     * Constructor for the AmURLConnection
     */
    protected AmHttpURLConnection( URL url ) {
        super( url );
    }

    /**
     * Open the local file
     */
    public void connect() throws IOException {
        if ( ! connected ) {
            int port = url.getPort();
            if ( port < 0 ) {
                port = 80;
            }
            Socket socket = new Socket( url.getHost(), port );
            instream = socket.getInputStream();
            String request = "GET "+url+"\n";
            int len = request.length();
            byte bytes[] = new byte[len];
            request.getBytes( 0, len, bytes, 0 );
            socket.getOutputStream().write( bytes );
        }
    }

    /**
     * Return the open stream to the local file (open if necessary).
	 *
	 * @return instream		the open stream
     */
    public InputStream getInputStream() throws IOException {
        if ( ! connected ) {
            connect();
        }
        if ( instream == null ) {
            throw new IOException();
        }
        return instream;
    }

    /**
     * Return the content type, if known, or null.
	 *
	 * @return		String
     */
    public String getContentType() {
        if ( url.getFile().toLowerCase().endsWith(".gif") ) {
            return "image/gif";
        }
        else if ( url.getFile().toLowerCase().endsWith(".jpg") ) {
            return "image/jpg";
        }
        else {
            return null;
        }
    }
}
