/****************************************************************
 **
 **  $Id: TextLine.java,v 1.31 1998/07/11 20:23:48 kadel Exp $
 **
 **  $Source: /cvs/classes/dtai/gwt/TextLine.java,v $
 **
 ****************************************************************
 **
 **  Gadget Windowing Toolkit (GWT) Java Class Library
 **  Copyright (C) 1997  DTAI, Incorporated (http://www.dtai.com)
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Library General Public
 **  License as published by the Free Software Foundation; either
 **  version 2 of the License, or (at your option) any later version.
 **
 **  This library is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 **  Library General Public License for more details.
 **
 **  You should have received a copy of the GNU Library General Public
 **  License along with this library (file "COPYING.LIB"); if not,
 **  write to the Free Software Foundation, Inc.,
 **  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 **
 ****************************************************************/

package com.dtai.gwt;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.util.StringTokenizer;

/**
 * TextLine
 * @version 1.1
 * @author DTAI, Incorporated
 */
class TextLine extends DisplayGadget {

    StringBuffer text;
    Color overrideForeground;
    private boolean editable = true;
    private boolean cursorAlwaysOn = false;
    private int columns;
    private int drawXOffset = 0;
    private char echoChar = '\0';
    int selectionStart = 0;
    int selectionEnd = 0;
    int caretPosition = 0;
    boolean mouseDown = false;
    private boolean showCursor = false;
    private StringBuffer undoText;
    private int undoSelectionStart = 0;
    private int undoSelectionEnd = 0;
    private int undoCaretPosition = 0;
    private boolean startUndo = true;

    private boolean lineBreak = false;
    private Rectangle selectionRect;
    private int cursorOffset;
    private int cursorWidth = 1;

    /**
     * TextLine
     */
    public TextLine() {
        this("");
    }

    /**
     * TextLine
     * @param text
     */
    public TextLine(String text) {
        setBackground(Gadget.transparent);
		setHorizAlign( HORIZ_ALIGN_LEFT );
        setText(text);
        setCursor(GadgetCursor.getPredefinedCursor(GadgetCursor.TEXT_CURSOR));
    }
    
	private boolean charIsWord(char ch) {
	    return ((font instanceof SpecialFont) &&
        		(((SpecialFont)font).charIsWord(ch)));
    }

    /**
     * gets the preffered size
     * @return Dimension
     */
    public Dimension getPreferredSize() {
	    FontMetrics metrics = getFontMetrics(getFont());
	    String str = text.toString();
	    pref_width = metrics.stringWidth(str)+2;
        if (metrics instanceof SpecialFontMetrics) {
            pref_height = ((SpecialFontMetrics)metrics).stringAscent(str) +
                          metrics.getDescent() + metrics.getLeading();
        } else {
    	    pref_height = metrics.getHeight();
        }
        return new Dimension( pref_width, pref_height );
    }

	/**
	 * gets the text length
	 * @return int
	 */
	public int getTextLength() {
		return text.length();
	}

	/**
	 * gets the text of the button
	 * @return String
	 */
	public String getText() {
		return text.toString();
	}

	/**
	 * toString
	 * @return String
	 */
	public String toString() {
	    return text.toString();
	}

	/**
	 * setText
	 * @param text
	 */
	public void setText( String text ) {
		synchronized(getTreeLock()) {

    	    if ( startUndo ) {
    	        saveUndo();
    	    }
    	    if ( text == null ) {
    	        text = "";
    	        select( 0 );
    	    }
    	    if ( ( this.text == null ) ||
    	         ( ! this.text.toString().equals( text ) ) ) {
                this.text = new StringBuffer(text.length());
                StringTokenizer st = new StringTokenizer( text, "\t\n\r" );
                while ( st.hasMoreTokens() ) {
                    this.text.append( st.nextToken() );
                    if ( st.hasMoreTokens() ) {
                        this.text.append( " " );
                    }
                }
                select( 0 );
        		repaint();
        	}
        }
	}

    /**
     * Returns the boolean indicating whether this TextComponent is editable or not.
     * @see setEditable
     * @return boolean
     */
    public final boolean isEditable() {
        return editable;
    }

    /**
     * Sets the specified boolean to indicate whether or not this TextComponent should be editable.
     * @param  b - the boolean to be set
     * @see isEditable
     */
    public final void setEditable(boolean b) {
        editable = b;
        if (editable) {
            setCursor(GadgetCursor.getPredefinedCursor(GadgetCursor.TEXT_CURSOR));
        } else {
            setCursor(null);
        }
    }

    /**
     * Returns the boolean indicating whether this TextComponent is cursorAlwaysOn or not.
     * @see setCursorAlwaysOn
     * @return boolean
     */
    public final boolean isCursorAlwaysOn() {
        return cursorAlwaysOn;
    }

    /**
     * Sets the specified boolean to indicate whether or not this TextComponent should be cursorAlwaysOn.
     * @param  b - the boolean to be set
     * @see isCursorAlwaysOn
     */
    public final void setCursorAlwaysOn(boolean b) {
        cursorAlwaysOn = b;
    }

    /**
     * Returns the cursor width
     * @return the current width (defaults to 1)
     */
    public final int getCursorWidth() {
        return cursorWidth;
    }

    /**
     * Sets the cursor width (defaults to 1)
     * @param  cursorWidth the new width
     */
    public final void setCursorWidth(int cursorWidth) {
        if (cursorWidth != this.cursorWidth) {
            this.cursorWidth = cursorWidth;
            repaint();
        }
    }

    /**
     * insertText
     * @param str
     * @param pos
     */
    public void insertText(String str, int pos) {
        if (pos < 0) {
            pos = Math.max(0,text.length()-(pos+1));
        }
        text.insert( pos, str );
    }

    /**
     * insertText
     * @param c
     * @param pos
     */
    public void insertText(char c, int pos) {
        text.insert( caretPosition, c );
    }

    /**
     * appendText
     * @param str
     */
    public void appendText(String str) {
        text.append(str);
    }

    /**
     * appendText
     * @param c
     */
    public void appendText(char c) {
        text.append(c);
    }

    /**
     * replaceText
     * @param str
     * @param start
     * @param end
     */
    public void replaceText(String str, int start, int end) {
        String string = text.toString();
	    text = new StringBuffer( string.substring(0, start) );
	    text.append( str );
	    text.append( string.substring(end) );
    }

    /**
     * replaceText
     * @param c
     * @param start
     * @param end
     */
    public void replaceText(char c, int start, int end) {
        String string = text.toString();
	    text = new StringBuffer( string.substring(0, start) );
	    text.append( c );
	    text.append( string.substring(end) );
    }

	/**
	 * getShowCursor
	 * @return boolean
	 */
	public boolean getShowCursor() {
		return showCursor;
	}

	/**
	 * setShowCursor
	 * @param showCursor
	 */
	public void setShowCursor( boolean showCursor ) {
	    this.showCursor = showCursor;
	    repaint();
	}

	/**
	 * getMouseDown
	 * @return boolean
	 */
	public boolean getMouseDown() {
		return mouseDown;
	}

	/**
	 * setMouseDown
	 * @param mouseDown
	 */
	public void setMouseDown( boolean mouseDown ) {
	    this.mouseDown = mouseDown;
	}

	/**
	 * toggleCursor
	 */
	public void toggleCursor() {
	    setShowCursor( ! showCursor );
	}

	/**
	 * getColumns
	 * @return int
	 */
	public int getColumns() {
		return columns;
	}

	/**
	 * isTextSelected
	 * @return boolean
	 */
	public boolean isTextSelected() {
	    if (selectionStart == selectionEnd) {
	        return false;
	    } else {
	        return true;
	    }
	}

	/**
	 * setColumns
	 * @param columns
	 */
	public void setColumns( int columns ) {
		synchronized(getTreeLock()) {
    	    this.columns = columns;
    	    Dimension min_size = getMinimumSize();

    	    if ( ( width != min_size.width ) ||
    	         ( height != min_size.height ) ) {
    	        invalidate();
    	    }
    	}
	}

    /**
     * getEchoChar
     * @return char
     */
    public char getEchoChar() {
        return echoChar;
    }

    /**
     * setEchoChar
     * @param c
     */
    public void setEchoChar(char c) {
        echoChar = c;
        repaint();
    }

    /**
     * echoCharIsSet
     * @return boolean
     */
    public boolean echoCharIsSet() {
        return ( echoChar > '\0' );
    }

	/**
	 * saveUndo
	 */
	protected void saveUndo() {
        undoText = text;
        undoSelectionStart = selectionStart;
        undoSelectionEnd = selectionEnd;
        undoCaretPosition = caretPosition;
        startUndo = false;
	}

	/**
	 * undo
	 */
	protected void undo() {
	    if ( undoText != null ) {
	        StringBuffer tempText = text;
            int tempSelectionStart = selectionStart;
            int tempSelectionEnd = selectionEnd;
            int tempCaretPosition = caretPosition;

            text = undoText;
            selectionStart = undoSelectionStart;
            selectionEnd = undoSelectionEnd;
            caretPosition = undoCaretPosition;

            undoText = tempText;
            undoSelectionStart = tempSelectionStart;
            undoSelectionEnd = tempSelectionEnd;
            undoCaretPosition = tempCaretPosition;
            repaint();
        }
	}

	/**
	 * getSelectBg
	 * @param g
	 * @param fg
	 * @return Color
	 */
	protected Color getSelectBg(GadgetGraphics g, Color fg) {
        Color selectionColor = getSelectedBackground(g);
        if ( selectionColor == null ) {
            return fg;
        }
        else {
            return selectionColor;
        }
	}

	private Color getSelectFg(GadgetGraphics g) {
        Color selectedTextColor = getSelectedForeground(g);
        if ( selectedTextColor == null ) {
            return getNormalBackground(g);
        }
        else {
            return selectedTextColor;
        }
    }

    private void paintSelection( GadgetGraphics g, Color fg, int x, int y, int text_y,
	                             String draw_string,
	                             int text_height ) {

		if ( selectionStart < selectionEnd ) {
		    FontMetrics metrics = getFontMetrics(getFont());
		    int start_x = x + metrics.stringWidth(
                    draw_string.substring( 0, selectionStart ) );
		    int end_x = x + metrics.stringWidth(
                    draw_string.substring( 0, selectionEnd ) );

            g.setColor( getSelectBg(g,fg) );
            selectionRect = new Rectangle(start_x, y, ( end_x - start_x ), text_height);
            g.fillRect(selectionRect.x,selectionRect.y,selectionRect.width,selectionRect.height);

            g.setColor( getSelectFg(g) );
			g.drawString( draw_string.substring( selectionStart, selectionEnd ), start_x, text_y,
			    y, y+text_height );
		}
	}
	
	public int getCursorOffset() {
	    return cursorOffset;
	}

	public Rectangle getSelectionRect() {
	    if (selectionRect == null) {
	        return new Rectangle(0, 0, 0, height);
	    }
	    return new Rectangle(selectionRect.x,selectionRect.y,selectionRect.width,selectionRect.height);
	}

    /**
     * paintDisabledForeground
     * @param g
     * @param x
     * @param y
     * @param width
     * @param height
     */
    protected void paintDisabledForeground( GadgetGraphics g,
                                    int x, int y, int width, int height) {
	    paintForeground(g,x,y,width,height);
	}

    /**
     * paints the foreground
     * @param g
     * @param x
     * @param y
     * @param width
     * @param height
     */
    protected void paintForeground( GadgetGraphics g,
                                    int x, int y, int width, int height) {
        x=0;
        Color fg;
        if ( overrideForeground == null ) {
            fg = getForeground(g);
        }
        else {
            fg = overrideForeground;
        }
        if (!editable && parent instanceof TextFieldGadget &&
            parent.parent instanceof ChoiceGadget && hasFocus()) {
            g.setColor(getSelectBg(g,fg));
            g.fillRect(0,0,width,height);
            fg = getSelectFg(g);
        }
		if ( fg != null ) {
			String draw_string;

			if ( echoCharIsSet() ) {
			    StringBuffer echo = new StringBuffer( text.length() );
			    for ( int i = 0; i < text.length(); i++ ) {
			        echo.append( echoChar );
			    }
			    draw_string = echo.toString();
			}
			else {
			    draw_string = getText();
			}

			FontMetrics metrics = g.getFontMetrics();
			int ascent;
            if (metrics instanceof SpecialFontMetrics) {
                ascent = ((SpecialFontMetrics)metrics).stringAscent(draw_string);
            } else {
                ascent = metrics.getAscent();
            }
			int text_y = y + ascent;
			int text_height = ascent + metrics.getDescent() + metrics.getLeading();

            int visibleWidth = getSize().width;

        	int string_width = metrics.stringWidth( draw_string );
 			int pixels_to_cursor = metrics.stringWidth( draw_string.substring( 0, caretPosition ) );

    	    int align = getHorizAlign();

        	if ( ( visibleWidth <= 0 ) ||
        	     ( string_width + 2 < visibleWidth ) ) {
        	    drawXOffset = 0;
        	}
        	else {
        	    if ( hasFocus() ) {
        	        align = HORIZ_ALIGN_LEFT;
        	    }

    			int offset_pixels_to_cursor = drawXOffset + pixels_to_cursor;

    			while ( offset_pixels_to_cursor < 0 ) {
    			    drawXOffset += visibleWidth / 5;
    			    if ( drawXOffset > 0 ) {
    			        drawXOffset = 0;
    			    }
			        offset_pixels_to_cursor = drawXOffset + pixels_to_cursor;
    			}
    			while ( offset_pixels_to_cursor > visibleWidth ) {
    			    drawXOffset -= visibleWidth / 5;
    			    if ( drawXOffset < ( visibleWidth - string_width - 1 ) ) {
    			        drawXOffset = visibleWidth - string_width - 1;
    			    }
			        offset_pixels_to_cursor = drawXOffset + pixels_to_cursor;
    			}
            }

            if ( align != HORIZ_ALIGN_LEFT ) {
                drawXOffset = visibleWidth - ( string_width + 2 );
            }
            if ( align == HORIZ_ALIGN_CENTER ) {
                drawXOffset /= 2;
            }

			x += drawXOffset;

			g.setColor(fg);
			g.drawString( draw_string, x + 1 /* room for cursor */, text_y, y, y+height );

			if ( hasFocus() || cursorAlwaysOn) {
    			selectionRect = null;
    			paintSelection( g, fg, x + 1 /* room for cursor */, y, text_y,
    			                draw_string, text_height );

			    int cursor_x = x + 1 + pixels_to_cursor;
                cursorOffset = cursor_x;
    			if ( showCursor && editable ) {
        			g.setColor( fg );
        			for (int i = 1; i <= cursorWidth; i++) {
        			    cursor_x += (i/2) * (((i%2) == 0) ? 1 : -1);
        			    g.drawLine( cursor_x, y, cursor_x, y + text_height - 1 );
        			}
    			}
			    if (selectionRect == null) {
                    selectionRect = new Rectangle(cursorOffset, y, 1, text_height);
                }
    		}
		}
	}

	/**
	 * findTextPos
	 * @param x
	 * @return int
	 */
	protected int findTextPos( int x ) {
        // for now, assume we have not scrolled at all
        FontMetrics metrics = getFontMetrics( getFont() );
        x -= drawXOffset;
        if ( x < 0 ) {
            return 0;
        }
        String str = getText();
        int prev_x = 0;
        int i;
        for ( i = 0; i < str.length(); i++ ) {
            int next_x = metrics.stringWidth( str.substring( 0, i+1 ) );
            if ( x < ( next_x - ( ( next_x - prev_x ) / 2 ) ) ) {
                break;
            }
            prev_x = next_x;
        }
        return i;
	}

	/**
	 * findTextPos
	 * @param x
	 * @return int
	 */
	protected int findCharPos( int x ) {
        // for now, assume we have not scrolled at all
        FontMetrics metrics = getFontMetrics( getFont() );
        x -= drawXOffset;
        if ( x < 0 ) {
            return 0;
        }
        String str = getText();
        int prev_x = 0;
        int i;
        int len = str.length();
        for ( i = 0; i < len; i++ ) {
            int next_x = metrics.stringWidth( str.substring( 0, i+1 ) );
            if ( x < next_x ) {
                break;
            }
            prev_x = next_x;
        }
        return Math.min(len-1, i);
	}

	/**
	 * findCaretPosition
	 * @param pos
	 * @return int
	 */
	protected int findCaretPosition( int pos ) {
        FontMetrics metrics = getFontMetrics( getFont() );
        String str = getText();
        if (pos > length()) {
            pos = length();
        }
        return metrics.stringWidth( str.substring( 0, pos ) );
	}

	/**
	 * getForegroundSize
     * @param g
     * @return Dimension
     */
	protected Dimension getForegroundSize( GadgetGraphics g ) {
	    return getForegroundSize( g, columns );
	}

	/**
	 * getForegroundSize
	 * @param g
	 * @param columns
	 * @return Dimension
	 */
	protected Dimension getForegroundSize( GadgetGraphics g, int columns ) {
		FontMetrics metrics = getFontMetrics(getFont(g));
		Dimension fg = new Dimension();
	    String str = text.toString();
        if (metrics instanceof SpecialFontMetrics) {
            fg.height = ((SpecialFontMetrics)metrics).stringAscent(str) +
                          metrics.getDescent() + metrics.getLeading();
        } else {
    		fg.height = metrics.getHeight();
        }
		if (columns == 0) {
		    fg.width = metrics.stringWidth(getText());
		}
		else {
		    int maxwidth = metrics.stringWidth( "W" );
    		fg.width = (columns * maxwidth);
        }
    	fg.width += 2; // the 1-pixel-wide cursor has to fit
		               // at beginning or end of the string.
		return fg;
	}

	/**
	 * resets the foreground size
	 * @param g
	 */
	protected void resetForegroundSize( GadgetGraphics g ) {
	    Dimension fgsize = getForegroundSize(g);
		fgHeight = fgsize.height;
		fgWidth = width;
	}

    /**
     * selectFromCursor
     * @param pos
     */
    protected void selectFromCursor(int pos) {
		if ( caretPosition == selectionStart ) {
			select( selectionEnd, pos );
		}
		else {
			select( selectionStart, pos );
		}
	}

	/**
	 * deleteSelection
	 */
	protected void deleteSelection() {
		//synchronized(getTreeLock()) {
    	    if ( selectionStart < selectionEnd ) {
        	    if ( startUndo ) {
        	        saveUndo();
        	    }
    	        String str = getText();
    	        text = new StringBuffer();
    	        if ( selectionStart > 0 ) {
    	            text.append( str.substring( 0, selectionStart ) );
    	        }
    	        if ( selectionEnd < str.length() ) {
    	            text.append( str.substring( selectionEnd, str.length() ) );
    	        }

    	        selectionEnd = selectionStart;
    	        caretPosition = selectionStart;
    	        startUndo = true;
    	    }
    	//}
	}

    /**
     * backspaceKey
     */
    protected void backspaceKey() {
		synchronized(getTreeLock()) {
            if ( ( selectionStart == selectionEnd ) &&
                 ( selectionStart > 0 ) ) {
                selectionStart--;
            }
            deleteSelection();
        }
    }

    /**
     * deleteKey
     */
    protected void deleteKey() {
		synchronized(getTreeLock()) {
    	    if ( startUndo ) {
    	        saveUndo();
    	    }
            if ( ( selectionStart == selectionEnd ) &&
                 ( selectionEnd < text.length() ) ) {
                selectionEnd++;
            }
            deleteSelection();
        }
    }

	/**
	 * typeChar
     * @param c
     */
	protected void typeChar( char c ) {
		synchronized(getTreeLock()) {
    	    if ( startUndo ) {
    	        saveUndo();
    	    }
    	    deleteSelection();
            insertText(c,caretPosition);
    	    select( caretPosition+1 );
    	    repaint();
    	}
	}

	/**
	 * cut
	 */
	protected void cut() {
		synchronized(getTreeLock()) {
            GadgetClipboard.setText( getSelectedText() );
            deleteSelection();
            repaint();
        }
	}

	/**
	 * copy
	 */
	protected void copy() {
		synchronized(getTreeLock()) {
            GadgetClipboard.setText( getSelectedText() );
        }
	}

	/**
	 * paste
	 */
	protected void paste() {
		synchronized(getTreeLock()) {
    	    if ( startUndo ) {
    	        saveUndo();
    	    }
            deleteSelection();
            String toPaste = GadgetClipboard.getText();
            text.insert( caretPosition, toPaste );
            select( selectionStart+toPaste.length() );
            repaint();
        }
	}

	/**
	 * moveTo
	 * @param nextPos
	 * @param shiftDown
	 */
	protected void moveTo( int nextPos, boolean shiftDown ) {
		synchronized(getTreeLock()) {
            if ( nextPos >= 0 ) {
        	    if ( shiftDown ) {
        	        if ( caretPosition == selectionStart ) {
        	            select( selectionEnd, nextPos );
        	        }
        	        else {
        	            select( selectionStart, nextPos );
        	        }
                }
                else {
    	            select( nextPos );
    	        }
    	    }
    	}
	}

	/**
	 * moveLeft
	 * @param ctrlDown
	 * @param shiftDown
	 */
	protected void moveLeft( boolean ctrlDown, boolean shiftDown ) {
		synchronized(getTreeLock()) {
    	    int nextPos = -1;
            if ( ! ctrlDown ) {
                nextPos = caretPosition - 1;
            }
            else {
                nextPos = caretPosition;
                String str = getText();
                char ch;
                if ( ( nextPos == str.length() ) ||
                     ( (Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos--;
                }
                while ( ( nextPos >= 0 ) &&
                        ( (Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos--;
                }
                while ( ( nextPos >= 0 ) &&
                        ( !(Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos--;
                }
                if ( ( nextPos >= 0 ) &&
                     ( (Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos++;
                }
            }
            if ( nextPos < 0 ) {
                nextPos = 0;
            }
            moveTo( nextPos, shiftDown );
        }
    }

	/**
	 * moveRight
	 * @param ctrlDown
	 * @param shiftDown
	 */
	protected void moveRight( boolean ctrlDown, boolean shiftDown ) {
		synchronized(getTreeLock()) {
    	    int nextPos = -1;
            if ( ! ctrlDown ) {
                nextPos = caretPosition + 1;
            }
            else {
                nextPos = caretPosition;
                String str = getText();
                char ch;
                while ( ( nextPos < str.length() ) &&
                        ( (Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos++;
                }
                while ( ( nextPos < str.length() ) &&
                        ( !(Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos++;
                }
                while ( ( nextPos < str.length() ) &&
                        ( (Character.isSpace(ch = str.charAt( nextPos )) || charIsWord(ch)) ) ) {
                    nextPos++;
                }
            }
            if ( nextPos > text.length() ) {
                nextPos = text.length();
            }
            moveTo( nextPos, shiftDown );
        }
	}

    /**
     * Returns the selected text's start position.
     * @return int
     */
    public int getSelectionStart() {
        return selectionStart;
    }

    /**
     * Sets the selection start to the specified position.
	 * @param selectionStart - the start position of the text
	 * The new starting point is constrained to be before or at the current selection end.
     */
    public void setSelectionStart(int selectionStart) {
		synchronized(getTreeLock()) {
            if ( selectionStart < 0 ) {
                selectionStart = 0;
            }
            else if ( selectionStart > text.length() ) {
                selectionStart = text.length();
            }
            this.selectionStart = selectionStart;
            caretPosition = selectionStart;
	        startUndo = true;
            repaint();
        }
    }

    /**
     * Returns the selected text's end position.
     * @return int
     */
    public int getCaretPosition() {
        return caretPosition;
    }

    /**
     * Sets the selected text's end position.
     * @param caretPosition
     */
    public void setCaretPosition(int caretPosition) {
        if (caretPosition != this.caretPosition) {
            this.caretPosition = caretPosition;
	        startUndo = true;
            repaint();
        }
    }

    /**
     * Returns the selected text's end position.
     * @return int
     */
    public int getSelectionEnd() {
        return selectionEnd;
    }

    /**
     * Sets the selection end to the specified position. The new end point is constrained to be at or after the current selection start.
     * @param selectionEnd - the start position of the text
     */
    public void setSelectionEnd(int selectionEnd) {
		synchronized(getTreeLock()) {
            if ( selectionEnd < 0 ) {
                selectionEnd = 0;
            }
            else if ( selectionEnd > text.length() ) {
                selectionEnd = text.length();
            }
            this.selectionEnd = selectionEnd;
            caretPosition = selectionEnd;
	        startUndo = true;
            repaint();
        }
    }

    /**
     * Selects the text found between the specified start and end locations.
     * @param selectionStart - the start position of the text
     * @param selectionEnd - the end position of the text
     */
    public void select( int selectionStart,
                                     int selectionEnd ) {
		synchronized(getTreeLock()) {
            if ( selectionStart < 0 ) {
                selectionStart = 0;
            }
            else if ( selectionStart > text.length() ) {
                selectionStart = text.length();
            }
            if ( selectionEnd < 0 ) {
                selectionEnd = 0;
            }
            else if ( selectionEnd > text.length() ) {
                selectionEnd = text.length();
            }
            if ( selectionStart > selectionEnd ) {
                this.selectionStart = selectionEnd;
                this.selectionEnd = selectionStart;
            }
            else {
                this.selectionStart = selectionStart;
                this.selectionEnd = selectionEnd;
            }
            caretPosition = selectionEnd;
	        startUndo = true;
            repaint();
        }
    }

    /**
     * select
     * @param pos
     */
    public void select( int pos ) {
        select( pos, pos );
    }

    /**
     * Selects all the text in the TextComponent.
     */
    public void selectAll() {
        select( 0, text.length() );
    }

    /**
     * Selects all the text in the TextComponent.
     * @return int
     */
    public int length() {
        return text.length();
    }

    /**
     * Returns the selected text contained in this TextComponent.
     * @see setText
     * @return String
     */
    public String getSelectedText() {
		synchronized(getTreeLock()) {
            return getText().substring( selectionStart, selectionEnd );
        }
    }

    /**
     * sets if there is a carrage return at line of line
     * @param in
     */
    public void setLineBreak (boolean in) {
        lineBreak = in;
    }

    /**
     * returns true if the line had a carrage return at end
     * @return boolean
     */
    public boolean getLineBreak ( ) {
        return lineBreak;
    }

}
