package org.signwriting;

import com.dtai.gwt.GadgetGraphics;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Toolkit;

public class SignCharacter extends SignComponent {
    
    static final long serialVersionUID = -4015900762071756890L;

    private static final FontMetrics metrics = SignFont.large.getMetrics();
    private static final int fontHeight = metrics.getHeight();

    private String stringForm;

    public SignCharacter(byte ascii, byte y, byte x) {
    	byte buf[] = {ascii};
        stringForm = new String(buf,0);
        this.y = y*2;
        this.x = ((PCNumbers.toUnsigned(x) - 8 ) / 4) * getWidth();
    }
    
    protected int calcWidth() {
        return metrics.stringWidth(stringForm);
    }
    
    protected int calcHeight() {
        return fontHeight;
    }

    private SignCharacter(SignCharacter rhs) {
        super(rhs);
        stringForm = rhs.stringForm;
    }

    public Object clone() {
        SignCharacter newSignCharacter = new SignCharacter(this);
        return newSignCharacter;
    }

    public int hashCode() {
        return stringForm.hashCode();
    }

    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        SignCharacter rhs = (SignCharacter)obj;
        if (stringForm.charAt(0) != rhs.stringForm.charAt(0)) {
            return false;
        }
        return true;
    }

    public void paint(GadgetGraphics g, int x, int y, int offsetx, int offsety,
                      Bitmap bitmap, int[] pixels, PaintInfo paintInfo) {
        if (paintInfo != null) {
            paintInfo.valid = false;
        }
        if (bitmap != null) {
            bitmap.setValid(false); // leave this here until I put in a
                                    // better way to indicate this when
                                    // passing in pixel array
        }
        boolean thumbNail = false;
        Font font = g.getFont();
        thumbNail = (font == SignFont.small);
        if (thumbNail) {
            x += ((this.x) / 2);
            y += ((this.y) / 2);
        } else {
            x += this.x;
            y += this.y;
        }
        y += Toolkit.getDefaultToolkit().getFontMetrics(font).getAscent();
        g.drawString(stringForm,x+offsetx,y-2+offsety);
    }

    public String toString() {
        return "Alphabet: "+stringForm+" y="+y+" x="+x;
    }
}
