/****************************************************************
 **
 **  $Id: SignDisplay.java,v 1.11 1998/09/17 02:09:32 ccp Exp $
 **
 **  $Source: /cvs/classes/signwriting/SignDisplay.java,v $
 **
 ****************************************************************/

package org.signwriting;

import com.dtai.gwt.CanvasGadget;
import com.dtai.gwt.GadgetGraphics;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Event;
import java.awt.Insets;
import java.awt.Rectangle;
import java.util.Vector;
import java11.awt.event.AWTEvent;
import java11.awt.event.FocusEvent;
import java11.awt.event.InputEvent;
import java11.awt.event.KeyEvent;
import java11.awt.event.MouseEvent;

/**
 * SignDisplay
 * @version 1.1
 * @author DTAI, Incorporated
 */
public class SignDisplay extends CanvasGadget implements Runnable {

    private static final int NONE = -1;
    private static final int RIGHT = 0;
    private static final int LOWER_RIGHT = 1;
    private static final int DOWN = 2;
    private static final int LOWER_LEFT = 3;
    private static final int LEFT = 4;
    private static final int UPPER_LEFT = 5;
    private static final int UP = 6;
    private static final int UPPER_RIGHT = 7;

    private static final int MARGIN = 16;

    Sign sign;
    int signX;
    int signY;
    Point cursorPos;
    int displayWidth;
    int displayHeight;
    SignComponent cursorComponent;
    int cursorType;
    boolean cursorVisible;
    Thread cursorThread;
    boolean fixedSize = false;
    boolean editable = false;
    boolean newCursor = false;
    KeyBoxes keyboxes;
    private Point dragOrigin = new Point();
    private SignComponent[] dragComponents;
    private Point[] dragComponentOrigins;
    private Sign copySign;
    private SignWriter signwriter;
    
    /**
     * Constructs a SignDisplay.
     */
	public SignDisplay(SignWriter signwriter) {
	    this.signwriter = signwriter;
	    clear();
	}
	
	public SignDisplay() {
	    this(null);
	}

	public void setFixedSize(boolean fixedSize) {
	    this.fixedSize = fixedSize;
	}

	public boolean isEditable() {
	    return editable;
	}

	public void setEditable(boolean editable) {
	    this.editable = editable;
	}

	public void setKeyBoxes(KeyBoxes keyboxes) {
	    this.keyboxes = keyboxes;
	}

    public void deleteSelectedComponents() {
	    SignComponent[] selected = sign.getSelectedComponents();
	    for (int i = 0; i < selected.length; i++) {
	        sign.remove(selected[i]);
	    }
	    repaint();
    }
    
    public void copySelectedComponents() {
	    SignComponent[] selected = sign.getSelectedComponents();
	    copySign = new Sign();
	    for (int i = 0; i < selected.length; i++) {
	        copySign.add((SignComponent)selected[i].clone());
	    }
    }
    
    public void cutSelectedComponents() {
        copySelectedComponents();
        deleteSelectedComponents();
    }
    
    public void paste() {
        if (copySign != null) {
            sign.deselectAllComponents();
    	    SignComponent[] copyComponents = copySign.getComponents();
    	    for (int i = 0; i < copyComponents.length; i++) {
    	        SignComponent clone = (SignComponent)copyComponents[i].clone();
    	        sign.add(clone);
    	        clone.setSelected(true);
    	    }
    	    repaint();
        }
    }
    
    public void duplicateSelectedComponents() {
        copySelectedComponents();
        paste();
    }
    
    /**
     * clears the box.
     */
	public void clear() {
	    sign = null;
	    hideCursor();
	    displayWidth = 75;
	    displayHeight = 75;
	    signX = 0;
	    signY = 0;
	    setSize(0,0);
	}

	public void showCursor() {
	    Sign sign = getSign();
	    if (editable && sign != null) {
	        Rectangle signBounds = sign.getBounds();
            cursorPos = new Point(signX+signBounds.width, signY+signBounds.height);
            if (cursorThread == null) {
                cursorType = RIGHT;
                cursorThread = new Thread(this);
                cursorThread.start();
            }
            cursorVisible = true;
        }
	}

	/**
	 * run the thread to blink the cursor
	 */
	public void run() {
	    Thread currentThread = Thread.currentThread();
	    while (cursorThread == currentThread) {
	        try {
    	        currentThread.sleep(500);
    	    } catch (InterruptedException ie) {
    	    }
    	    if (cursorThread == currentThread) {
    	        cursorVisible = !cursorVisible;
    	        repaint();
    	    }
	    }
	}

	public void processMotion(int xChange, int yChange) {
	    Sign sign = this.sign;
	    if (sign != null) {
            if (getFont() == SignFont.small) {
                xChange *= 2;
                yChange *= 2;
            }
    	    SignComponent[] selected = sign.getSelectedComponents();
    	    if (selected.length > 0) {
        	    for (int i = 0; i < selected.length; i++) {
        	        selected[i].x += xChange;
        	        selected[i].y += yChange;
        	    }
    		    signChanged();
    		} else {
    		    moveCursor(xChange,yChange);
    		}
	    }
    }

	/**
	 * move the cursor by the given x and y offsets
	 */
	public void moveCursor(int xChange, int yChange) {
	    if (cursorPos != null) {
	        cursorPos.x += xChange;
	        cursorPos.y += yChange;
	        newCursor = true;
    	    repaint();
	    }
	}

	/**
	 * rotate (and show if necessary) the cursor, around the last symbol
	 */
	public void rotateCursor(boolean clockwise) {
	    if (sign != null) {
	        if (cursorType == NONE) {
	            SignComponent[] selected = sign.getSelectedComponents();
	            if (selected.length == 1) {
	                cursorComponent = selected[0];
	                cursorThread = null;
	            }
	        }
    	    sign.deselectAllComponents();
    	    if (cursorType == NONE) {
    	        cursorType = RIGHT;
    	    } else if (clockwise) {
    	        cursorType += 1;
    	        cursorType %= 8;
    	    } else {
    	        cursorType -= 1;
    	        if (cursorType < 0) {
    	            cursorType = 7;
    	        }
    	    }
            newCursor = false;
    	    if (cursorComponent != null || cursorPos != null) {
                cursorVisible = true;
                repaint();
    	        if (cursorThread == null) {
	                cursorThread = new Thread(this);
	                cursorThread.start();
	            }
	        }
    	}
	}

	/**
	 * hide the cursor
	 */
	public void hideCursor() {
	    cursorThread = null;
	    cursorComponent = null;
	    cursorPos = null;
	    cursorVisible = false;
	    cursorType = NONE;
	    repaint();
	}

    /**
     * adds a symbol to the box
     */
	public Sign getSign() {
	    return sign;
	}

    /**
     * returns the offset based on cursor position
     */
	private Point getCursorOffset(int width, int height) {
	    if (cursorPos == null) {
	        return new Point(0, 0);
	    }
	    Point offset = new Point(cursorPos.x, cursorPos.y);

        switch(cursorType) {
            case DOWN:
                offset.x -= width/2;
                break;
            case LOWER_LEFT:
                offset.x -= width - 1;
                break;
            case LEFT:
                offset.x -= width - 1;
                offset.y -= height/2;
                break;
            case UPPER_LEFT:
                offset.x -= width - 1;
                offset.y -= height - 1;
                break;
            case UP:
                offset.x -= width/2;
                offset.y -= height - 1;
                break;
            case RIGHT:
                offset.y -= height/2;
                break;
            case UPPER_RIGHT:
                offset.y -= height - 1;
                break;
            case LOWER_RIGHT:
                break;
            default:
                throw new IllegalArgumentException("Bad cursor type: "+cursorType);
        }
	    return offset;
	}

   /**
     * adds a symbol to the box
     */
	public void addSignComponent(SignComponent signComponent) {
        Point offset = getCursorOffset(signComponent.getWidth(),
                                       signComponent.getHeight());
        addSignComponent(signComponent,offset.x,offset.y);
	}

   /**
     * adds a symbol to the box
     */
	public void addSignComponent(SignComponent signComponent, int xOffset, int yOffset) {

        signComponent.x += xOffset;
        signComponent.y += yOffset;

	    if (sign == null) {
	        sign = new Sign();
	    }
	    sign.add(signComponent);
	    signChanged();
	}

   /**
     * sets the symbol to display in the box.
     */
	public void setSignComponent(SignComponent signComponent) {
	    clear();
	    addSignComponent(signComponent);
	    signChanged();
	}

    /**
     * sets the sign to display in the box.
     */
	public void setSign(Sign sign) {
	    this.sign = sign;
	    signChanged();
	}

    /**
     * adds the sign to the box.
     */
	public void addSign(Sign sign) {

	    Rectangle signBounds = sign.getBounds();

        Point offset = getCursorOffset(signBounds.width,signBounds.height);
	    SignComponent[] components = sign.getComponents();

	    for (int i = 0; i < components.length; i++) {
	        addSignComponent(components[i],offset.x,offset.y);
	    }
	}

	/**
     * something about the sign changed, so repaint and adjust size if necessary
     */
	protected void signChanged() {
	    if (!fixedSize) {
    	    Dimension min = getMinimumSize();
            if (getFont() == SignFont.small) {
                min.width *= 2;
                min.height *= 2;
            }
    	    int displayWidth = this.displayWidth;
    	    int displayHeight = this.displayHeight;
    	    if (min.width+MARGIN > displayWidth || min.height+MARGIN > displayHeight) {
    	        int newWidth = displayWidth;
    	        int newHeight = displayHeight;

        	    if (min.width+MARGIN > displayWidth) {
        	        newWidth = min.width+MARGIN;
        	    }
        	    if (min.height+MARGIN > displayHeight) {
        	        newHeight = min.height+MARGIN;
        	    }
    	        setPreferredSize(newWidth,newHeight);
        	    setSize(0,0);//cause invalidate, and overlay panel will know to resize
    	    }
    	}
	    repaint();
	}

	/**
     * grow the displayed area
     */
	public void grow() {
	    Point p = getLocation();
	    Dimension d = signwriter.textArea.getSize();
        if (p.x + displayWidth + 20 <= d.width && 
            p.y + displayHeight + 20 <= d.height) {
	        setPreferredSize(displayWidth+20, displayHeight+20);
	    }
	}

	/**
     * shrink the displayed area
     */
	public void shrink() {
        if (displayWidth  - 20 >= 75 && 
            displayHeight - 20 >= 75) {
	        setPreferredSize(displayWidth-20, displayHeight-20);
	    }
	}

	/**
     * crop the displayed area to the minimum size
     */
	public void crop() {
	    setPreferredSize(0,0);//will get reset to minimum
	}

	/**
     * change the preferred size as long as it's not smaller than the minimum
     */
	public void setPreferredSize(int width, int height) {
	    if (!fixedSize) {
    	    Dimension min = getMinimumSize();
            if (getFont() == SignFont.small) {
                min.width *= 2;
                min.height *= 2;
            }
    	    displayWidth = Math.max(min.width,width)+MARGIN;
    	    displayHeight = Math.max(min.height,height)+MARGIN;
    	    setSize(0,0);
    	    repaint();
    	}
	}

	/**
	 * getPreferredSize
	 * @return Dimension
	 */
	public Dimension getPreferredSize() {
	    if (fixedSize) {
    	    int pw = pref_width;
    	    int ph = pref_height;
    	    if (pw < 0 || ph < 0) {
    	        pw = 50;
    	        ph = 50;
                if (getFont() == SignFont.small) {
                    pw /= 2;
                    ph /= 2;
                }
    	        pref_width = pw;
    	        pref_height = ph;
    	    }
        	return new Dimension(pw,ph);
        } else {
            pref_width = displayWidth;
            pref_height = displayHeight;
            if (getFont() == SignFont.small) {
                pref_width /= 2;
                pref_height /= 2;
            }
        	return new Dimension(pref_width,pref_height);
        }
	}

	/**
	 * getMinimumSize
	 * @return Dimension
	 */
	public Dimension getMinimumSize() {
	    int mw = 0;
	    int mh = 0;
        if (sign != null) {
            Rectangle signBounds = sign.getBounds();
            mw = signBounds.width;
            mh = signBounds.height;
            if (getFont() == SignFont.small) {
                mw /= 2;
                mh /= 2;
            }
        }
    	return new Dimension(mw,mh);
	}

    /**
     * Paint the box
     */
    public void paint(GadgetGraphics g) {
        if (sign != null) {
            Dimension size = getSize();
            
            if (dragComponents == null) {

                Rectangle signBounds = sign.getBounds();
                if (getFont() == SignFont.small) {
                    signBounds.width++;
                    signBounds.height++;
                    signBounds.width /= 2;
                    signBounds.height /= 2;
                }

                signX = (size.width-signBounds.width)/2-signBounds.x;
                signY = (size.height-signBounds.height)/2-signBounds.y;
            }

            //puts a cyan or pink rectanlgle behind the signs in the 
            //sign edit box so you can see where the edit box is.
            if (signwriter != null && signwriter.textArea != null) {
                Graphics gg = g.getOrigGraphics();
                Point p = getLocation();
                Color c = gg.getColor();
                Color paper = signwriter.textArea.getTextBackground();
                if (paper.equals(Color.cyan)) {
                    gg.setColor(Color.pink);
                }
                else {
                    gg.setColor(Color.cyan);
                }
                gg.fillRect(p.x+1, p.y+1, size.width-2, size.height-2);
                gg.setColor(c);
            }
            
            sign.paint(g, signX, signY);
            if (cursorVisible && sign.getComponentCount() > 0) {
                if (cursorComponent == null) {
                    cursorComponent = sign.getComponent(0);
                }
                if (!newCursor) {
                    calculateCursorPos(cursorComponent, cursorType);
                }
                g.setXORMode(getBackground(g));
                paintCursor(g,
                    cursorComponent, signX,
                    signY, cursorType);
                g.setPaintMode();
            }
        }
    }

    public void setVisible(boolean visible, boolean invalidateParent) {
        if (visible != isVisible()) {
            if (visible && signwriter != null && (!(getParent() instanceof KeyBox))) {
                setBackground(signwriter.getPaperColor());
                setForeground(signwriter.getInkColor());
            }
            super.setVisible(visible, invalidateParent);
            if (!visible && keyboxes != null) {
                keyboxes.invokeAction(null);
            }
        }
    }

    /**
     * processMouseEvent
     * @param mouse the MouseEvent
     */
    public void processMouseMotionEvent( MouseEvent e ) {
        if (editable && dragComponents != null) {
            if (e.getID() == MouseEvent.MOUSE_DRAGGED) {
                for (int i = 0; i < dragComponents.length; i++) {
                    int newx = dragComponentOrigins[i].x+(e.getX()-dragOrigin.x);
                    int newy = dragComponentOrigins[i].y+(e.getY()-dragOrigin.y);
                    if (getFont() == SignFont.small) {
                        newx *= 2;
                        newy *= 2;
                    }
                    dragComponents[i].setLocation(newx, newy);
                }
        	    repaint();
            }
        }
    }

    /**
     * processMouseEvent
     * @param mouse the MouseEvent
     */
    public void processMouseEvent( MouseEvent e ) {
        if (editable) {
            if (e.getID() == MouseEvent.MOUSE_RELEASED) {
                if (dragComponents != null) {
                    signChanged();
                }
            }
            if (e.getID() == MouseEvent.MOUSE_PRESSED ||
                e.getID() == MouseEvent.MOUSE_RELEASED) {
                dragComponents = null;
                e.consume();
            }
            if (e.getID() == MouseEvent.MOUSE_PRESSED) {
                int xpos = e.getX();
                int ypos = e.getY();
                int localSignX = signX;
                int localSignY = signY;
                if (getFont() == SignFont.small) {
                    xpos *= 2;
                    ypos *= 2;
                    localSignX *= 2;
                    localSignY *= 2;
                }
                /*
                if (keyboxes != null) {
                    keyboxes.requestFocus();
                }*/
                signwriter.textArea.requestFocus();
                if (sign != null) {
                    SignComponent component = sign.getComponentAt(xpos-localSignX,
                                                                  ypos-localSignY);
                    if (component != null) {
                        if (!e.isShiftDown() && !component.isSelected()) {
                            sign.deselectAllComponents();
                        }
                        component.setSelected(true);
                        dragComponents = sign.getSelectedComponents();
                        sign.toFront(component);
                        dragComponentOrigins = new Point[dragComponents.length];
                        for (int i = 0; i < dragComponents.length; i++) {
                            dragComponentOrigins[i] = dragComponents[i].getLocation();
                            if (getFont() == SignFont.small) {
                                dragComponentOrigins[i].x /= 2;
                                dragComponentOrigins[i].y /= 2;  
                            }
                        }
                        dragOrigin = e.getPoint();
                        hideCursor();
                    } else {
                        if (!e.isShiftDown()) {
                            sign.deselectAllComponents();
                        }
                        cursorPos = new Point(xpos-localSignX, ypos-localSignY);
            	        newCursor = true;
                        if (cursorThread == null) {
                            cursorType = UPPER_RIGHT;
                            cursorThread = new Thread(this);
                            cursorThread.start();
                        }
                        cursorVisible = true;
                    }
                    repaint();
                }
            }
        }
	    super.processMouseEvent( e );
	}

    public void calculateCursorPos(SignComponent comp, int cursorType) {
        
        final int gap = 2;
        
        int xoffset = comp.x;
        int yoffset = comp.y;
        
        int width = comp.getWidth();
        int height = comp.getHeight();
        
        switch(cursorType) {
            case RIGHT:
                cursorPos = new Point(xoffset+width+gap,
                    yoffset+(int)(height/2));
                break;
            case LOWER_RIGHT:
                cursorPos = new Point(xoffset+width+gap,
                    yoffset+height+gap);
                break;
            case DOWN:
                cursorPos = new Point(xoffset+(int)(width/2),
                    yoffset+height+gap);
                break;
            case LOWER_LEFT:
                cursorPos = new Point(xoffset-1-gap,
                    yoffset+height+gap);
                break;
            case LEFT:
                cursorPos = new Point(xoffset-1-gap,
                    yoffset+(int)(height/2));
                break;
            case UPPER_LEFT:
                cursorPos = new Point(xoffset-1-gap,
                    yoffset-1-gap);
                break;
            case UP:
                cursorPos = new Point(xoffset+(int)(width/2),
                    yoffset-1-gap);
                break;
            case UPPER_RIGHT:
                cursorPos = new Point(xoffset+width+gap, yoffset-1-gap);
                break;
            default:
                throw new IllegalArgumentException("Bad cursor type: "+cursorType);
         }
    }

    public void paintCursor(GadgetGraphics g, SignComponent comp,
                             int x, int y, int cursorType) {
        int cursorX = cursorPos.x;
        int cursorY = cursorPos.y;
        if (g.getFont() == SignFont.small) {
            cursorX++;
            cursorY++;
            cursorX /= 2;
            cursorY /= 2;
        }
        switch(cursorType) {
            case RIGHT:
                drawRightCursor(g, x+cursorX, y+cursorY);
                break;
            case LOWER_RIGHT:
                drawLowerRightCursor(g, x+cursorX, y+cursorY);
                break;
            case DOWN:
                drawDownCursor(g, x+cursorX, y+cursorY);
                break;
            case LOWER_LEFT:
                drawLowerLeftCursor(g, x+cursorX, y+cursorY);
                break;
            case LEFT:
                drawLeftCursor(g, x+cursorX, y+cursorY);
                break;
            case UPPER_LEFT:
                drawUpperLeftCursor(g, x+cursorX, y+cursorY);
                break;
            case UP:
                drawUpCursor(g, x+cursorX, y+cursorY);
                break;
            case UPPER_RIGHT:
                drawUpperRightCursor(g, x+cursorX, y+cursorY);
                break;
            default:
                throw new IllegalArgumentException("Bad cursor type: "+cursorType);
         }
    }

    public void drawLowerRightCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y,x,y+4);
            //g.drawLine(x+1,y,x+1,y+4);
            g.drawLine(x+1,y,x+4,y);
            //g.drawLine(x,y+1,x+4,y+1);
        } else {
            g.drawLine(x,y,x,y+8);
            g.drawLine(x+1,y,x+1,y+8);
            g.drawLine(x+2,y,x+8,y);
            g.drawLine(x+2,y+1,x+8,y+1);
        }
    }

    public void drawDownCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x-3/*7*/,y,x+4,y);
            //g.drawLine(x-3/*7*/,y+1,x+4,y+1);
            g.drawLine(x,y+1,x,y+3);
            //g.drawLine(x+1,y,x+1,y+3);
        } else {
            g.drawLine(x-7,y,x+8,y);
            g.drawLine(x-7,y+1,x+8,y+1);
            g.drawLine(x,y+2,x,y+6);
            g.drawLine(x+1,y+2,x+1,y+6);
        }
    }

    public void drawLeftCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y-3/*7*/,x,y+4);
            //g.drawLine(x-1,y-3/*7*/,x-1,y+4);
            g.drawLine(x-1,y,x-3,y);
            //g.drawLine(x,y+1,x-3,y+1);
        } else {
            g.drawLine(x,y-7,x,y+8);
            g.drawLine(x-1,y-7,x-1,y+8);
            g.drawLine(x-2,y,x-6,y);
            g.drawLine(x-2,y+1,x-6,y+1);
        }
    }

    public void drawRightCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y-3/*7*/,x,y+4);
            //g.drawLine(x+1,y-3/*7*/,x+1,y+4);
            g.drawLine(x+1,y,x+3,y);
            //g.drawLine(x,y+1,x+3,y+1);
        } else {
            g.drawLine(x,y-7,x,y+8);
            g.drawLine(x+1,y-7,x+1,y+8);
            g.drawLine(x+2,y,x+6,y);
            g.drawLine(x+2,y+1,x+6,y+1);
        }
    }

    public void drawLowerLeftCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y,x,y+4);
            //g.drawLine(x-1,y,x-1,y+4);
            g.drawLine(x-1,y,x-4,y);
            //g.drawLine(x,y+1,x-4,y+1);
        } else {
            g.drawLine(x,y,x,y+8);
            g.drawLine(x-1,y,x-1,y+8);
            g.drawLine(x-2,y,x-8,y);
            g.drawLine(x-2,y+1,x-8,y+1);
        }
    }

    public void drawUpperLeftCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y,x,y-4);
            //g.drawLine(x-1,y,x-1,y-4);
            g.drawLine(x-1,y,x-4,y);
            //g.drawLine(x,y-1,x-4,y-1);
        } else {
            g.drawLine(x,y,x,y-8);
            g.drawLine(x-1,y,x-1,y-8);
            g.drawLine(x-2,y,x-8,y);
            g.drawLine(x-2,y-1,x-8,y-1);
        }
    }

    public void drawUpCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x-3/*7*/,y,x+4,y);
            //g.drawLine(x-3/*7*/,y-1,x+4,y-1);
            g.drawLine(x,y-1,x,y-3);
            //g.drawLine(x+1,y,x+1,y-3);
        } else {
            g.drawLine(x-7,y,x+8,y);
            g.drawLine(x-7,y-1,x+8,y-1);
            g.drawLine(x,y-2,x,y-6);
            g.drawLine(x+1,y-2,x+1,y-6);
        }
    }

    public void drawUpperRightCursor(GadgetGraphics g, int x, int y) {
        if (g.getFont() == SignFont.small) {
            g.drawLine(x,y,x,y-4);
            //g.drawLine(x+1,y,x+1,y-4);
            g.drawLine(x+1,y,x+4,y);
            //g.drawLine(x,y-1,x+4,y-1);
        } else {
            g.drawLine(x,y,x,y-8);
            g.drawLine(x+1,y,x+1,y-8);
            g.drawLine(x+2,y,x+8,y);
            g.drawLine(x+2,y-1,x+8,y-1);
        }
    }
}
