package org.signwriting;

import com.dtai.gwt.Gadget;
import com.dtai.gwt.AppletManager;
import com.dtai.gwt.ButtonGadget;
import com.dtai.gwt.CheckboxGadget;
import com.dtai.gwt.CheckboxGadgetGroup;
import com.dtai.gwt.BorderGadget;
import com.dtai.gwt.GadgetApplet;
import com.dtai.gwt.GadgetBorderLayout;
import com.dtai.gwt.GadgetFlowLayout;
import com.dtai.gwt.GadgetGridLayout;
import com.dtai.gwt.GadgetGraphics;
import com.dtai.gwt.CanvasGadget;
import com.dtai.gwt.TextAreaGadget;
import com.dtai.gwt.PanelGadget;
import com.dtai.gwt.LabelGadget;
import com.dtai.gwt.MaskedTextFieldGadget;
import com.dtai.gwt.MenuBarGadget;
import com.dtai.gwt.MenuGadget;
import com.dtai.gwt.MenuItemGadget;
import com.dtai.gui.SimpleDialog;
import com.dtai.gwt.ScrollbarGadget;
import com.dtai.util.ShowUser;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Frame;
import java.awt.FileDialog;
import java.io.IOException;
import java.io.StreamCorruptedException;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.ObjectOutputStream;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.ObjectInputStream;
import java.io.File;
import java11.awt.event.ActionEvent;
import java11.awt.event.ActionListener;
import java11.awt.event.KeyEvent;
import java11.awt.event.KeyListener;
import java11.awt.event.ItemEvent;
import java11.awt.event.ItemListener;
import java11.awt.event.AdjustmentEvent;
import java11.awt.event.AdjustmentListener;
import java11.awt.event.FocusEvent;
import java11.awt.event.FocusListener;
import java.util.Vector;

//PrintJob
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.PrintJob;
import com.dtai.gwt.ContainerGadget;

//Loading window
import java.awt.Window;


public class SignWriter extends GadgetApplet {

    public static final byte[] NEW_SGNFILE_MAGIC =
        {(byte)0xDE, (byte)0xAF, (byte)0xCA, (byte)'N'};
    public static final String DEFAULT_TITLE = "SignWriter 5.0";

    private static final int YES = 1;
    private static final int NO = 2;
    private static final int CANCEL = 3;

    public final static int SIGN_LANGUAGE = 0;
    public final static int WORDS = 1;
    public final static int BOTH = 2;
    public final static int FINGERSPELLING = 3;
    /* Should have SIGNS, FINGERSPELLING, WORDS for what mode we are in (which
    then checks what language has been chosen and loads the appropriate KeyFile,
    Dictionary, etc), and three more flags like MENUS_IN_SIGNS, MENUS_IN_WORDS,
    MENUS_IN_BOTH for the menu choices pickable in ViewPrefs */
    
    public static StringBuffer allSymbolsText;
    public static int testStart;
    public static int testEnd;

    private ViewPrefsDialog viewPrefsDialog;
    private DictionaryDialog dictionaryDialog;
    MyTextArea textArea;
    private SignDisplay display;
    private KeyBoxes keyBoxes;
    private MenuItemGadget exit;
    private FileDialog openDialog;
    private FileDialog saveDialog;
    private CloseDialog closeDialog;
    private boolean madeChanges = false;
    private LabelGadget loadingLabel;
    private Vector uiModeListeners = new Vector();
    private int language = SIGN_LANGUAGE;
    CheckboxGadget signModeCb;
    CheckboxGadget fingerModeCb;
    CheckboxGadget textModeCb;
    private AboutDialog aboutDialog;

    /*
     * The current filename
     */
    private String filename;

    public static void main(String args[]) {
        String newFilename = null;
        if (args.length > 0) {
            if (args[0].equals("-test")) {
                allSymbolsText = new StringBuffer();
                if (args.length < 3) {
                    System.err.println("args: -test 0 10 (0 is start code, 10 is end code)");
                    System.exit(1);
                } else {
                    testStart = Integer.parseInt(args[1]);
                    testEnd = Integer.parseInt(args[2]);
                }
            } else if (!args[0].startsWith("-")) {
                newFilename = args[0];
                if (!newFilename.toLowerCase().endsWith(".sgn")) {
                    newFilename += ".sgn";
                }
            }
        }
        AppletManager manager = new AppletManager(new SignWriter(newFilename),
                                                  700, 500, args);
    }

    public SignWriter(String filename) {
        this.filename = filename;
        Resources.openResources(this);
    }

    private Window loadingWindow;

    public void start() {
        String rememberFilename = filename;
        setFilename(null);
        filename = rememberFilename;

        getMainPanel().setEnabled(false);
        getMainPanel().getShell().showWaitCursor();

        setLayout(new GadgetBorderLayout());
        loadingLabel = new LabelGadget("Please wait...");
        loadingLabel.setFont(new Font("SansSerif", Font.BOLD, 18));
        add("Center", loadingLabel);
        loadingLabel.getParent().repaint(); // GWT Bug
        new LoaderThread(this).start();
    }

    void load() {
        synchronized (getMainPanel().getTreeLock()) {
            try {
                loadingLabel.setLabel("Loading Symbols. Please wait...");
                new SymbolFile().readSymbols(loadingLabel);
                loadingLabel.setLabel("Loading keyboard information. Please wait...");
                new KeyFile().readKeys();
            } catch (IOException ioe) {
                ShowUser.error(ioe,"File error!");
            }
            remove(loadingLabel);
            display = new SignDisplay(SignWriter.this);
            display.setVisible(false);
            keyBoxes = new KeyBoxes(display);
            textArea = new MyTextArea(SignWriter.this, keyBoxes);
            add("Center", textArea);
            setEditFont(SignFont.large);
            readConfiguration();
            PanelGadget signEdit = new PanelGadget();
            signEdit.setLayout(new GadgetBorderLayout());
            signEdit.add("Center", keyBoxes);
            BorderGadget mode = new BorderGadget();
            //mode.setBorderType(BorderGadget.ETCHED_IN);
            //mode.setBorderThickness(2);
            mode.setLayout(new GadgetGridLayout(0,3));
            CheckboxGadgetGroup modeGroup = new CheckboxGadgetGroup();
            GadgetFlowLayout flow = new GadgetFlowLayout(GadgetFlowLayout.CENTER, 0, 0);
            PanelGadget panel = new PanelGadget();
            panel.setLayout(flow);
            mode.add(panel);
            panel.add(signModeCb = new NoFocusCheckboxGadget("Type Signs",true,modeGroup));
            panel = new PanelGadget();
            panel.setLayout(flow);
            mode.add(panel);
            panel.add(fingerModeCb = new NoFocusCheckboxGadget("Type Fingerspelling",false,modeGroup));
            panel = new PanelGadget();
            panel.setLayout(flow);
            mode.add(panel);
            panel.add(textModeCb = new NoFocusCheckboxGadget("Type Words",false,modeGroup));
            signModeCb.addItemListener(
                new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        textArea.setMode(SIGN_LANGUAGE);
                    }
                }
            );
            fingerModeCb.addItemListener(
                new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        textArea.setMode(FINGERSPELLING);
                    }
                }
            );
            textModeCb.addItemListener(
                new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        textArea.setMode(WORDS);
                    }
                }
            );
            /*
            signModeCb.addItemListener(itemListener);
            fingerModeCb.addItemListener(itemListener);
            textModeCb.addItemListener(itemListener);
            */
            signEdit.add("North", mode);
            add("South", signEdit);
            MenuBarGadget menubar = new EditorMenuBar(SignWriter.this);
            add("North",menubar);
            textArea.getShell().setDefaultFocusGadget(textArea);
            if (allSymbolsText != null) {
                textArea.setText(allSymbolsText.toString());
            } else if (filename != null) {
                open(filename);
            }
            getMainPanel().setEnabled(true);
            getMainPanel().getShell().hideWaitCursor();
        }
    }

	/**
	 * Adds the specified listener to be notified when component
	 * events occur on this component.
	 *
	 * @param l 	the listener to receive the events
	 */
	public void addUIModeListener(UIModeListener l) {
        uiModeListeners.addElement(l);
	}

	/**
	 * Removes the specified listener so it no longer receives
	 * cell events on this cell.
	 *
	 * @param l 		the listener to remove
	 */
	public void removeUIModeListener(UIModeListener l) {
        uiModeListeners.removeElement(l);
	}

	public void fireUIModeEvent() {
	    processUIModeEvent(new UIModeEvent(this, language));
	}

	protected void processUIModeEvent(UIModeEvent e) {
	    UIModeListener[] listeners = new UIModeListener[uiModeListeners.size()];
	    uiModeListeners.copyInto(listeners);
	    for (int i = 0; i < listeners.length; i++) {
			listeners[i].uiModeChanged(e);
		}
	}

    public void setEditFont(Font font) {
        display.setFont(font);
        keyBoxes.setFont(font);
        textArea.setFont(font);
    }

    public Font getEditFont() {
        return display.getFont();
    }

    public Color getWindowColor() {
        return getMainPanel().getBackground();
    }

    public void setWindowColor(Color color) {
        getMainPanel().setBackground(color);
        java.awt.Container parent = getMainPanel().getShell();
        while (parent != null) {
            parent.setBackground(color);
            parent = parent.getParent();
        }
    }

    public Color getPaperColor() {
        return textArea.getTextBackground();
    }

    public void setPaperColor(Color color) {
        textArea.setTextBackground(color);
    }

    public Color getInkColor() {
        return textArea.getForeground();
    }

    public void setInkColor(Color color) {
        textArea.setForeground(color);
    }

    public void setLanguage(int language) {
        if (this.language != language) {
            this.language = language;
        }
        fireUIModeEvent(); // Kludge...do this outside check for
                           // change so dialog colors get reset
    }

    public int getLanguage() {
        return language;
    }

    public void destroy() {
        try {
            closeDocument();
            writeConfiguration();
        } catch (CanceledOperationException coe) {
            throw new RuntimeException("Close canceled");
        }
        setFilename(null);
    }
    
    public void showAboutDialog() {
        if (aboutDialog == null) {
            aboutDialog = new AboutDialog(this);
        }
        aboutDialog.show();
    }

    private void closeDocument() throws CanceledOperationException {
        if (madeChanges) {
            if (closeDialog == null) {
                closeDialog = new CloseDialog(getFrame());
            }
            closeDialog.show();
            int answer = closeDialog.getAnswer();
            if (answer == CANCEL) {
                throw new CanceledOperationException();
            }
            if (answer == YES) {
                saveDocument();
            }
        }
        setFilename(null);
        textArea.setText("");
        madeChanges = false;
    }

    private final class CloseDialog extends SimpleDialog {

        private int answer;
        private ButtonGadget yes;
        private ButtonGadget no;
        private ButtonGadget cancel;

        CloseDialog(Frame frame) {
            super(frame, DEFAULT_TITLE, true, true);
            addButtonGadget(yes = new ButtonGadget("Yes"));
            yes.addActionListener(this);
            addButtonGadget(no = new ButtonGadget("No"));
            no.addActionListener(this);
            addButtonGadget(cancel = new ButtonGadget("Cancel"));
            cancel.addActionListener(this);
            yes.setAsDefault();
            setMessage("You have made changes to "+filename+".\n"+
                       "\n"+
                       "Do you want to save the changes?");
        }

        public void show() {
            answer = CANCEL;
            super.show();
        }

        int getAnswer() {
            return answer;
        }

    	/**
    	 * Checks if the user clicked on "OK"
    	 * @param action	the ActionEvent.
    	 */
        public void actionPerformed( ActionEvent action ) {

            Object button = action.getSource();
            if (button == yes) {
                answer = YES;
            } else if (button == no) {
                answer = NO;
            }
            hide();
        }
    }

    public void newDocument() {
        try {
            closeDocument();
        } catch (CanceledOperationException coe) {
        }
    }

    private FileDialog getOpenDialog() {
        if (openDialog == null) {
            openDialog = new FileDialog(getFrame(),
                "Open SignWriter File", FileDialog.LOAD);
        }
        return openDialog;
    }

    private FileDialog getSaveDialog() {
        if (saveDialog == null) {
            saveDialog = new FileDialog(getFrame(),
                "Save SignWriter File", FileDialog.SAVE);
        }
        return saveDialog;
    }
    
    public final static String sep=System.getProperty("path.separator");
 
    public void openDocument() {

        FileDialog open = getOpenDialog();
        open.setFile("*.sgn");
        open.show();
        String newFilename = open.getFile();
        String directory = open.getDirectory();
        if (newFilename != null) {
            if (!newFilename.endsWith(".sgn")) {
                newFilename += ".sgn";
            }
            open(directory+newFilename);
        }
    }

    private void open(String newFilename) {
        try {
            closeDocument();
            boolean done = false;
            FileInputStream filein = new FileInputStream(newFilename);
            if (filein.available() >= NEW_SGNFILE_MAGIC.length) {
                byte[] magicBytes = new byte[NEW_SGNFILE_MAGIC.length];
                filein.read(magicBytes);
                int i = 0;
                for (i = 0; i < magicBytes.length; i++) {
                    if (magicBytes[i] != NEW_SGNFILE_MAGIC[i]) {
                        break;
                    }
                }
                if (i == magicBytes.length) {
                    readNewSignFile(filein);
                    setFilename(newFilename);
                    done = true;
                }
            }
            filein.close();

            if (!done) {
                com.dtai.util.ShowUser.info(
                    "This file is being converted from the SignWriter 4.0 format.");

                try {
                    Thread.currentThread().sleep(2000); // wait for dialog to pop up
                } catch (InterruptedException e) {
                }

                DocumentPart[] docParts = new SWFile(newFilename).readDocument();

                char[] docBuffer = new char[docParts.length*2];
                int bufidx = 0;

                boolean nextSign = true;
                for (int i = 0; i < docParts.length; i++) {
                    if (docParts[i] == null) {
                        break;
                    }
                    if (docParts[i] == DocumentPart.NEW_LINE) {
                        docBuffer[bufidx++] = '\n';
                        nextSign = true;
                    } else {
                        nextSign = false;
                        Sign sign = (Sign)docParts[i];
                        SignFont.add(sign); // do this the first time you read the document in
                        docBuffer[bufidx++] = sign.getFontIndex();
                    }
                }
                textArea.setText(new String(docBuffer,0,bufidx));
                setFilename(null);
            }
            madeChanges = false;
        } catch (CanceledOperationException coe) {
        } catch (IOException ioe) {
            com.dtai.util.ShowUser.error(ioe,"Error reading SignWriter file "+newFilename);
        }
    }

    public void readNewSignFile(InputStream in) throws IOException {
        ObjectInputStream objin = new ObjectInputStream(
            new BufferedInputStream(in));
        try {
            String text = (String)objin.readObject();
            Sign[] signs = (Sign[])objin.readObject();
            char[] docBuffer = text.toCharArray();
            for (int i = 0; i < signs.length; i++) {
                if (signs[i] != null) {
                    SignFont.add(signs[i]);
                    docBuffer[i] = signs[i].getFontIndex();
                }
            }
            textArea.setText(new String(docBuffer,0,docBuffer.length));
            textArea.setCaretPosition(0);
        } catch (ClassNotFoundException cnfe) {
            cnfe.printStackTrace();
            throw new IOException("Bad serialized sign file.  "+cnfe.getMessage());
        } catch (StreamCorruptedException sce) {
            sce.printStackTrace();
            throw new IOException("Bad serialized sign file.  "+sce.getMessage());
        }
    }

    public void saveDocument() {
        saveDocument(filename);
    }

    public void saveDocument(String newFilename) {
        if (newFilename == null) {
            saveAsDocument();
            return;
        }
        try {
            BufferedOutputStream bufout = new BufferedOutputStream(
                new FileOutputStream(newFilename));
            bufout.write(NEW_SGNFILE_MAGIC);

            String text = textArea.getText();
            Sign[] signs = new Sign[text.length()];
            for (int i = 0; i < signs.length; i++) {
                char ch = text.charAt(i);
                if (ch >= SignFont.FONT_INDEX_START) {
                    signs[i] = SignFont.getSign(ch);
                }
            }

            ObjectOutputStream objout = new ObjectOutputStream(bufout);
            objout.writeObject(text);
            objout.writeObject(signs);
            objout.close();
            setFilename(newFilename);
        } catch (IOException ioe) {
            com.dtai.util.ShowUser.error(ioe,"Error writing SignWriter file "+newFilename);
        }
        madeChanges = false;
    }

    public void saveAsDocument() {
        FileDialog save = getSaveDialog();
        save.setFile("Untitled.sgn");
        save.show();
        String newFilename = save.getFile();
        boolean saveIt = true;
        //if newFilename is already in directory, warn user
        if (newFilename != null) {
            if (!newFilename.endsWith(".sgn")) {
                newFilename += ".sgn";
            }
            //try {
                //FileOutputStream fos = new FileOutputStream(newFilename, true);
                //fos.close();
                //we opened it without Exception, so it EXISTED, which is bad...
                File newFile=new File(newFilename);
                if(newFile.exists())
                { CloseDialog warning = new CloseDialog(getFrame());
                  warning.setMessage("The SignWriter file "+filename+" already\n"+
                  "exists.  Do you want to overwrite it?\n");
                  warning.show();
                  if (warning.getAnswer() != 1) {
                    saveIt = false;
                  }
                }
            //} catch (IOException ioe) {
            //    //it DID NOT EXIST, which is good in this case, so just save...
            //}
            if (saveIt) {
                saveDocument(newFilename);
            }
        }
    }

    public void setFilename(String newFilename) {
        filename = newFilename;
        if (filename == null) {
            getFrame().setTitle(DEFAULT_TITLE+" - Untitled");
        } else {
            getFrame().setTitle(DEFAULT_TITLE+" - "+filename);
        }
    }

    public void printDocument() 
    { Toolkit toolkit=Toolkit.getDefaultToolkit();
      PrintJob printJob=toolkit.getPrintJob(getFrame(),"",null);
      ContainerGadget container=(ContainerGadget)textArea.getLinesPanel();
      int panelHeight=container.getBounds().height;
      Dimension d=printJob.getPageDimension();
      for(int currHeight=0;currHeight<panelHeight;currHeight+=d.height)
      { Graphics g=printJob.getGraphics();
        Rectangle clip=new Rectangle(0,currHeight,d.width,d.height);
        GadgetGraphics gg=new GadgetGraphics(g,0,currHeight,d.width,d.height,container,true,true);
        container.printChildren(true,gg,clip); 
      }
      printJob.end();
    }

    public void exit() {
        try {
            closeDocument();
            writeConfiguration();
            System.exit(0);
        } catch (CanceledOperationException coe) {
        }
    }

    private void readConfiguration() {
        try {
            String configFileName = System.getProperty("user.home")+"/.sgnrc";
            System.err.print("Reading configuration file "+configFileName+"...");
            ObjectInputStream ios = new ObjectInputStream(
                new FileInputStream(configFileName));
            System.err.println("done");
            language = ((Integer)ios.readObject()).intValue();
            if (((Boolean)ios.readObject()).booleanValue()) {
                setEditFont(SignFont.large);
            } else {
                setEditFont(SignFont.small);
            }
            setWindowColor((Color)ios.readObject());
            setPaperColor((Color)ios.readObject());
            setInkColor((Color)ios.readObject());
        } catch (IOException e) {
        } catch (ClassNotFoundException e) {
        } catch (SecurityException e) {
        }
    }

    private void writeConfiguration() {
        try {
            String configFileName = System.getProperty("user.home")+"/.sgnrc";
            System.err.println("Writing configuration file "+configFileName);
            ObjectOutputStream oos = new ObjectOutputStream(
                new FileOutputStream(configFileName));
            oos.writeObject(new Integer(language));
            oos.writeObject(new Boolean(getEditFont() == SignFont.large));
            oos.writeObject(getWindowColor());
            oos.writeObject(getPaperColor());
            oos.writeObject(getInkColor());
            oos.close();
        } catch (IOException e) {
        } catch (SecurityException e) {
        }
    }

    public void undo() {
        textArea.undo();
    }

    public void cut() {
        textArea.cut();
    }

    public void copy() {
        textArea.copy();
    }

    public void paste() {
        textArea.paste();
    }

    public void delete() {
        insert("");
    }

    public void selectAll() {
        textArea.selectAll();
    }

    public void editViewPreferences() {
        if (viewPrefsDialog == null) {
            viewPrefsDialog = new ViewPrefsDialog(this);
        }
        viewPrefsDialog.show();
    }

    public void showDictionary() {
        if (dictionaryDialog == null) {
            dictionaryDialog = new DictionaryDialog(this);
        }
        dictionaryDialog.show();
    }

    public void insert(String text) {
        textArea.replaceRange(text,textArea.getSelectionStart(),textArea.getSelectionEnd());
    }

    private final class CanceledOperationException extends Exception {
        CanceledOperationException() {
            super();
        }
    }

    public final class MyTextArea extends TextAreaGadget implements KeyListener, ActionListener, FocusListener {

        SignWriter testApp;
        KeyBoxes keyboxes;
        int mode = SIGN_LANGUAGE;

        public MyTextArea(SignWriter testApp, KeyBoxes keyboxes) {
            this.testApp = testApp;
            this.keyboxes = keyboxes;
            addKeyListener(this);
            setShowGuides(true);
            setCursorWidth(3);
            setActiveUpdate(false);
            keyboxes.addActionListener(this);
            keyboxes.addFocusListener(this);
            setCursorAlwaysOn(true);
        }

        public boolean isFocusTraversable() 
        { //System.out.println("isFocusTraversable"); 
          //new Throwable().printStackTrace(System.out); 
          return true; 
        }
        
        public void focusGained(FocusEvent e) {
         //   setCursorAlwaysOn(true);
         System.out.println("textArea: focus Gained");
        }
        
        public void focusLost(FocusEvent e) {
         //   setCursorAlwaysOn(false);
         System.out.println("textArea: focus Lost");
        }
        
        public void setMode(int mode) {
            this.mode = mode;
            requestFocus();
        }

        public void fontChanged() {
            super.fontChanged();
        }

        public void validate() {
            super.validate();
        }
        
        //Tab key causes bad things to happen!!
        public void keyPressed(KeyEvent e) {
            if (!e.isActionKey()) {
                if ((e.getModifiers() & e.ALT_MASK) == e.ALT_MASK) {
                    //ALT KEY STUFF
                    switch(e.getKeyCode()) {
                    case 100:  //Alt-d - show Dictionary dialog
                    case 68:   //Alt-D 
                        showDictionary();
                        e.consume();
                        return;
                    case 102:  //Alt-f - change to Fingerspelling mode
                    case 70:   //Alt-F
                        fingerModeCb.setState(true);
                        e.consume();
                        return;
                    case 115:  //Alt-s - change to Sign mode
                    case 83:   //Alt-S
                        signModeCb.setState(true);
                        e.consume();
                        return;
                    case 119:  //Alt-w - change to Word (plain text) mode
                    case 87:   //Alt-W
                        textModeCb.setState(true);
                        e.consume();
                        return;                        
                        
                    case 110:  //Alt-n - move cursor to next line
                    case 78:   //Alt-N
                        moveDown(e.isControlDown(), e.isShiftDown());
                        e.consume();
                        return;
                    case 112:  //Alt-p - move cursor to previous line
                    case 80:   //Alt-P
                        moveUp(e.isControlDown(), e.isShiftDown());
                        e.consume();
                        return;
                        
                    case 108:  //Alt-l - jump to last sign in document,
                    case 76:   //Alt-L   ready to type AFTER that sign
                        setCaretPosition(getText().length() - 1);
                        scrollVisible();
                        e.consume();
                        return;
                    case 98:   //Alt-b - jump to beginning sign in document, 
                    case 66:   //Alt-B   ready to type BEFORE that sign
                        setCaretPosition(0);
                        scrollVisible();
                        e.consume();
                        return;
                  
                    //PENDING CASES:
                    
                    //PENDING:  MAKE Sure that ALT-cut, copy and paste work 
                    //on the selected symbol within a sign, if one is in the
                    //editing box...
                    case 120:  //Alt-x - cut a symbol within a sign while in 
                    case 88:   //Alt-X   the Editing Box
                        //PENDING: 
                        this.cut();
                        e.consume();
                        return;
                    case 99:   //Alt-c - copy a symbol within a sign while in
                    case 67:   //Alt-C   the Editing Box
                        //PENDING: 
                        this.copy();
                        e.consume();
                        return;
                    case 118:  //Alt-v - paste a symbol within a sign while
                    case 86:   //Alt-V   the Editing Box
                        //PENDING: 
                        this.paste();
                        e.consume();
                        return;
                    }
                } else if ((e.getModifiers() & e.CTRL_MASK) == e.CTRL_MASK) {
                    //CONTROL KEY STUFF
                    switch(e.getKeyCode()) {
                    case 14:    //Ctrl-n or Ctrl-N - new file
                        newDocument();
                        e.consume();
                        return;
                    case 15:    //Ctrl-o or Ctrl-O - open file
                        openDocument();
                        e.consume();
                        return;
                    case 16:    //Ctrl-p or Ctrl-P - print
                        printDocument();
                        e.consume();
                        return;
                    case 19:    //Ctrl-s or Ctrl-S - save file
                        saveDocument();
                        e.consume();
                        return;
                        
                    case 1:     //Ctrl-a or Ctrl-A - select all
                        this.selectAll();
                        e.consume();
                        return;
                    case 21:    //Ctrl-u or Ctrl-U - undo
                        this.undo();
                        e.consume();
                        return;
                        
                    //PENDING:  MAKE Sure that CTRL-cut, copy and paste work 
                    //on the complete sign, if one is in the editing box...
                    case 24:    //Ctrl-x or Ctrl-X - cut complete sign
                        this.cut();
                        e.consume();
                        return;
                    /*case 3:     //Ctrl-c or Ctrl-C - copy complete sign
                        this.copy();
                        e.consume();
                        return;*/
                    case 22:    //Ctrl-v or Ctrl-V - past complete sign
                        this.paste();
                        e.consume();
                        return;
                        
                    //PENDING CASES:
                    case 5:     //Ctrl-e or Ctrl-E 
                        //PENDING: exit file (?)
                        e.consume();
                        return;
                    case 17:    //Ctrl-q or Ctrl-Q
                        //PENDING: quit file (?)
                        e.consume();
                        return;
                        
                        
                    case 6:     //Ctrl-f or Ctrl-F
                        //PENDING: find
                        e.consume();
                        return;
                    case 18:    //Ctrl-r or Ctrl-R
                        //PENDING: replace
                        e.consume();
                        return;
                    case 7:     //Ctrl-g or Ctrl-G
                        //PENDING: find again
                        e.consume();
                        return;
                    }
                }
            }
            else {
                //e.isActionKey() == TRUE
                //HOME, END, LEFT, RIGHT, UP, and DOWN all have the
                //proper default behavior inherited from LinesPanel.java
                //we just need to handle PGUP, PGDN, and TAB...
                switch(e.getKeyCode()) {
                    case KeyEvent.PGUP:    //move cursor three lines up
                        moveUp(e.isControlDown(), e.isShiftDown(), 3);
                        e.consume();
                        return;
                    case KeyEvent.PGDN:    //move cursor three lines down
                        moveDown(e.isControlDown(), e.isShiftDown(), 3);
                        e.consume();
                        return;
                    /*case e.TAB:
                        //PENDING: move cursor from sign to sign
                        e.consume();
                        return;*/
                    /*case e.UP:      //move selected object up
                        e.consume();
                        return;*/
                }
            }
            madeChanges = true;
            if (mode == SignWriter.SIGN_LANGUAGE) {
                if (keyboxes.isActive()) {
                    keyboxes.processKeyEvent(e);
                } else if (!e.isActionKey()) {
                    if (e.getKeyChar() == e.ENTER) {
                        insert("\n");
                    } else {
                        editSign(e);
                    }
                }
            } else if (mode == SignWriter.FINGERSPELLING) {
    			int scanCode = KeyFile.getScanCode(e.getKeyChar());
                if (scanCode < 255 && scanCode != 14) {
        			InxRec entry =
        			    (InxRec)KeyFile.keyboard[KeyFile.FINGER_TYPE][scanCode][KeyFile.NO_MODIFIER];
            		int kinx = entry.inx;
                    int numsyms = KeyFile.keymap[kinx++];
        			Sign sign = new Sign();
        			sign.setWord(false);
                    for (int i = 0; i < numsyms; i++) {
                	    int symcode = PCNumbers.toUnsigned(KeyFile.keymap[kinx++]);
        		        int opt = KeyFile.keyToOpt(PCNumbers.toUnsigned(KeyFile.keymap[kinx++]));
                	    int y = KeyFile.keymap[kinx++];
                	    int x = KeyFile.keymap[kinx++];
                	    SignSymbol ss = new SignSymbol(symcode, opt, (byte)y, (byte)(x/2));
                        //hack fix for the letter 'z':
                        if (scanCode == 44 && numsyms == 2 && symcode == 107) {
                            ss.changeFlop();
                            ss.setRot(2);
                            ss.setFill(0);
                        }
//INITIALLY, Flop is set to TRUE on ALL Fingerspelling SignSymbols!
                        //in Fingerspelling, the [ and ] key are just flops
                        //of each other, so 27 has to remain flop=true:
            			if (scanCode != 27) {
            			    ss.changeFlop();
            			}
            			sign.add(ss);
            		}
        			SignFont.add(sign);
        			insert(""+sign.getFontIndex());
        			//insert(" ");
        			e.consume();
        		}
        		if (scanCode == 255) {
        		    //make space bar mean 3 spaces...
        		    insert("  ");
        		}
    		}
        }
        
        public void keyReleased(KeyEvent e) {
            if (mode == SignWriter.SIGN_LANGUAGE) {
                if (keyboxes.isActive()) {
                    keyboxes.processKeyEvent(e);
                }
            } else if (mode == SignWriter.FINGERSPELLING) {
    			e.consume();
    		}
        }
        
        public void keyTyped(KeyEvent e) {
            if (mode == SignWriter.SIGN_LANGUAGE) {
                if (keyboxes.isActive()) {
                    keyboxes.processKeyEvent(e);
                }
            } else if (mode == SignWriter.FINGERSPELLING) {
    			e.consume();
    		}
        }
        
        private void editSign(KeyEvent e) {
            String selected = getSelectedText();
            if (selected.length() > 1) {
                return;
            } else if (selected.length() == 0) {
                keyboxes.setSign(null);
            } else {
                Sign sign = SignFont.getSign(selected.charAt(0));
                if (sign != null) {
                    keyboxes.setSign(sign);
                }
            }
            Rectangle rect = getFirstSelectionRect();
            if (rect != null) {
                Point offset = getOffset();
                rect.x += offset.x;
                rect.y += offset.y;
                if (rect.width < 4) {
                    rect.x += ((SignFont)getFont()).horizontalMargin;
rect.x += 12; // TOTAL KLUDGE AND WON'T WORK FOR SMALL SIGNS
rect.y -= 3;
                }
                keyboxes.setDisplayLocation(rect);
            }
            if (!Character.isSpace(e.getKeyChar())) {
                keyboxes.processKeyEvent(e);
            }
            e.consume();
        }
        
        private void insert(String text) {
            int start = getSelectionStart();
            int end = getSelectionEnd();
            replaceRange(text,start,Math.min(end,start+1));
            select(start+1,start+1);
        }

        public void actionPerformed(ActionEvent e) {
            Sign sign = keyboxes.getSign();
            if (sign != null) {
                SignFont.add(sign);
                insert(""+sign.getFontIndex());
            }
            setEnabled(true);
            keyboxes.setSign(null);
            requestFocus();
        }

    }
    
    private static class NoFocusCheckboxGadget extends CheckboxGadget {
        public NoFocusCheckboxGadget(String label, boolean selected, CheckboxGadgetGroup group) {
            super(label, selected, group);
        }
    }
}

