
package org.signwriting;

import java.io.DataInputStream;
import java.io.InputStream;
import java.io.IOException;
import com.dtai.gwt.GadgetGraphics;

public class Symbol {

    final static int SGRID_SYMS = 1000;	/* Sym window symbol grid symbols  */
    private static Symbol[] table = new Symbol[SGRID_SYMS];
    private final static int	THUMB_SCALE = 2;		/* Thumbnail scale reduction  */

    private static final int IMAGE_CLASS1  = 0;
    private static final int IMAGE_CLASS2  = 1;
    private static final int IMAGE_CLASS3A = 2;
    private static final int IMAGE_CLASS3B = 3;
    private static final int IMAGE_CLASS4  = 4;
    private static final int IMAGE_CLASS5  = 5;
    private static final int IMAGE_CLASS6  = 6;
    private static final int IMAGE_CLASS7  = 7;

    private static final int NO_TRANSFORM = 0; // n  no image transform
    private static final int X_TRANSFORM  = 1; // X  invert image horizontally
    private static final int Y_TRANSFORM  = 2; // Y  invert image vertically
    private static final int XY_TRANSFORM = 3; // XY invert image horizontally and vertically

    private static ImageMap mapClass1      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(0,Y_TRANSFORM)  };
    private static ImageMap mapClass1Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(0,XY_TRANSFORM) };

    private static ImageMap mapClass2      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(0,Y_TRANSFORM)  };
    private static ImageMap mapClass2Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,X_TRANSFORM), new ImageMap(0,XY_TRANSFORM) };

    private static ImageMap mapClass3A     [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(1,Y_TRANSFORM), new ImageMap(0,Y_TRANSFORM), new ImageMap(1,XY_TRANSFORM),new ImageMap(2,X_TRANSFORM), new ImageMap(1,X_TRANSFORM)  };
    private static ImageMap mapClass3AFlop [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(1,XY_TRANSFORM),new ImageMap(0,XY_TRANSFORM),new ImageMap(1,Y_TRANSFORM), new ImageMap(2,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM) };

    private static ImageMap mapClass3B     [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(1,X_TRANSFORM), new ImageMap(0,X_TRANSFORM), new ImageMap(1,XY_TRANSFORM),new ImageMap(2,Y_TRANSFORM), new ImageMap(1,Y_TRANSFORM)  };
    private static ImageMap mapClass3BFlop [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(1,NO_TRANSFORM),new ImageMap(0,NO_TRANSFORM),new ImageMap(1,Y_TRANSFORM), new ImageMap(2,XY_TRANSFORM),new ImageMap(1,XY_TRANSFORM) };

    private static ImageMap mapClass4      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(3,NO_TRANSFORM),new ImageMap(0,XY_TRANSFORM),new ImageMap(1,XY_TRANSFORM),new ImageMap(2,XY_TRANSFORM),new ImageMap(3,XY_TRANSFORM) };
    private static ImageMap mapClass4Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(3,X_TRANSFORM), new ImageMap(0,Y_TRANSFORM), new ImageMap(1,Y_TRANSFORM), new ImageMap(2,Y_TRANSFORM), new ImageMap(3,Y_TRANSFORM)  };

    private static ImageMap mapClass5      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(3,NO_TRANSFORM),new ImageMap(4,NO_TRANSFORM),new ImageMap(3,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(1,X_TRANSFORM)  };
    private static ImageMap mapClass5Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(3,X_TRANSFORM), new ImageMap(4,X_TRANSFORM), new ImageMap(3,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM) };
   
    private static ImageMap mapClass6      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM),new ImageMap(3,NO_TRANSFORM) };
    private static ImageMap mapClass6Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(1,Y_TRANSFORM), new ImageMap(2,Y_TRANSFORM), new ImageMap(3,Y_TRANSFORM)  };

    private static ImageMap mapClass7      [] = { new ImageMap(0,NO_TRANSFORM),new ImageMap(1,NO_TRANSFORM),new ImageMap(2,NO_TRANSFORM) };
    private static ImageMap mapClass7Flop  [] = { new ImageMap(0,X_TRANSFORM), new ImageMap(2,X_TRANSFORM), new ImageMap(1,X_TRANSFORM)  };
    //  Map byte value into its bit-wise mirror image.
   
    private static byte flipByte[] = {
        (byte)0x00, (byte)0x80, (byte)0x40, (byte)0xC0, (byte)0x20, (byte)0xA0, (byte)0x60, (byte)0xE0,
        (byte)0x10, (byte)0x90, (byte)0x50, (byte)0xD0, (byte)0x30, (byte)0xB0, (byte)0x70, (byte)0xF0,

        (byte)0x08, (byte)0x88, (byte)0x48, (byte)0xC8, (byte)0x28, (byte)0xA8, (byte)0x68, (byte)0xE8,
        (byte)0x18, (byte)0x98, (byte)0x58, (byte)0xD8, (byte)0x38, (byte)0xB8, (byte)0x78, (byte)0xF8,

        (byte)0x04, (byte)0x84, (byte)0x44, (byte)0xC4, (byte)0x24, (byte)0xA4, (byte)0x64, (byte)0xE4,
        (byte)0x14, (byte)0x94, (byte)0x54, (byte)0xD4, (byte)0x34, (byte)0xB4, (byte)0x74, (byte)0xF4,

        (byte)0x0C, (byte)0x8C, (byte)0x4C, (byte)0xCC, (byte)0x2C, (byte)0xAC, (byte)0x6C, (byte)0xEC,
        (byte)0x1C, (byte)0x9C, (byte)0x5C, (byte)0xDC, (byte)0x3C, (byte)0xBC, (byte)0x7C, (byte)0xFC,

        (byte)0x02, (byte)0x82, (byte)0x42, (byte)0xC2, (byte)0x22, (byte)0xA2, (byte)0x62, (byte)0xE2,
        (byte)0x12, (byte)0x92, (byte)0x52, (byte)0xD2, (byte)0x32, (byte)0xB2, (byte)0x72, (byte)0xF2,

        (byte)0x0A, (byte)0x8A, (byte)0x4A, (byte)0xCA, (byte)0x2A, (byte)0xAA, (byte)0x6A, (byte)0xEA,
        (byte)0x1A, (byte)0x9A, (byte)0x5A, (byte)0xDA, (byte)0x3A, (byte)0xBA, (byte)0x7A, (byte)0xFA,

        (byte)0x06, (byte)0x86, (byte)0x46, (byte)0xC6, (byte)0x26, (byte)0xA6, (byte)0x66, (byte)0xE6,
        (byte)0x16, (byte)0x96, (byte)0x56, (byte)0xD6, (byte)0x36, (byte)0xB6, (byte)0x76, (byte)0xF6,

        (byte)0x0E, (byte)0x8E, (byte)0x4E, (byte)0xCE, (byte)0x2E, (byte)0xAE, (byte)0x6E, (byte)0xEE,
        (byte)0x1E, (byte)0x9E, (byte)0x5E, (byte)0xDE, (byte)0x3E, (byte)0xBE, (byte)0x7E, (byte)0xFE,

        (byte)0x01, (byte)0x81, (byte)0x41, (byte)0xC1, (byte)0x21, (byte)0xA1, (byte)0x61, (byte)0xE1,
        (byte)0x11, (byte)0x91, (byte)0x51, (byte)0xD1, (byte)0x31, (byte)0xB1, (byte)0x71, (byte)0xF1,

        (byte)0x09, (byte)0x89, (byte)0x49, (byte)0xC9, (byte)0x29, (byte)0xA9, (byte)0x69, (byte)0xE9,
        (byte)0x19, (byte)0x99, (byte)0x59, (byte)0xD9, (byte)0x39, (byte)0xB9, (byte)0x79, (byte)0xF9,

        (byte)0x05, (byte)0x85, (byte)0x45, (byte)0xC5, (byte)0x25, (byte)0xA5, (byte)0x65, (byte)0xE5,
        (byte)0x15, (byte)0x95, (byte)0x55, (byte)0xD5, (byte)0x35, (byte)0xB5, (byte)0x75, (byte)0xF5,

        (byte)0x0D, (byte)0x8D, (byte)0x4D, (byte)0xCD, (byte)0x2D, (byte)0xAD, (byte)0x6D, (byte)0xED,
        (byte)0x1D, (byte)0x9D, (byte)0x5D, (byte)0xDD, (byte)0x3D, (byte)0xBD, (byte)0x7D, (byte)0xFD,

        (byte)0x03, (byte)0x83, (byte)0x43, (byte)0xC3, (byte)0x23, (byte)0xA3, (byte)0x63, (byte)0xE3,
        (byte)0x13, (byte)0x93, (byte)0x53, (byte)0xD3, (byte)0x33, (byte)0xB3, (byte)0x73, (byte)0xF3,

        (byte)0x0B, (byte)0x8B, (byte)0x4B, (byte)0xCB, (byte)0x2B, (byte)0xAB, (byte)0x6B, (byte)0xEB,
        (byte)0x1B, (byte)0x9B, (byte)0x5B, (byte)0xDB, (byte)0x3B, (byte)0xBB, (byte)0x7B, (byte)0xFB,

        (byte)0x07, (byte)0x87, (byte)0x47, (byte)0xC7, (byte)0x27, (byte)0xA7, (byte)0x67, (byte)0xE7,
        (byte)0x17, (byte)0x97, (byte)0x57, (byte)0xD7, (byte)0x37, (byte)0xB7, (byte)0x77, (byte)0xF7,

        (byte)0x0F, (byte)0x8F, (byte)0x4F, (byte)0xCF, (byte)0x2F, (byte)0xAF, (byte)0x6F, (byte)0xEF,
        (byte)0x1F, (byte)0x9F, (byte)0x5F, (byte)0xDF, (byte)0x3F, (byte)0xBF, (byte)0x7F, (byte)0xFF
    };

    private int code;
    private byte type;
    private SymbolFill fills[][];

    public Symbol() {
    }

    private static void setSymbol(int code, Symbol symbol) {
        symbol.code = code;
        table[code] = symbol;
    }

    public static Symbol getSymbol(int code) {
        return table[code];
    }

    public final int getCode() {
        return code;
    }

    private int getCurrentOffset(int start,InputStream in) throws IOException {
        return start-in.available();
    }

    public void readSymbolData(int symCode, DataInputStream din) throws IOException {
        int start = din.available();
        int totalImageCount = 0;
        type = din.readByte();
        byte numVars = din.readByte();
        int[] varOffsets = new int[numVars];
        for (int i = 0; i < numVars; i++) {
            varOffsets[i] = din.readUnsignedShort();
        }

        fills = new SymbolFill[numVars][];
        for (int i = 0; i < numVars; i++) {
            if (getCurrentOffset(start,din) != varOffsets[i]) {
                throw new IOException("Error reading variation "+i+" at offset "+
                    getCurrentOffset(start,din)+"; was supposed to be at "+varOffsets[i]+
                    " so previous variation or symbol was bad");
            }
            int numFills = din.readByte();
            fills[i] = new SymbolFill[numFills];
            din.readByte(); // skip an unused byte
            int[] fillOffsets = new int[numFills];
            for (int j = 0; j < numFills; j++) {
                fillOffsets[j] = din.readUnsignedShort();
            }

            for (int j = 0; j < numFills; j++) {
                if (getCurrentOffset(start,din) != fillOffsets[j]) {
                    throw new IOException("Error reading variation "+i+", fill "+j+" at offset "+
                        getCurrentOffset(start,din)+"; was supposed to be at "+fillOffsets[j]+
                        " so previous fill, variation, or symbol was bad");
                }
                fills[i][j] = new SymbolFill();
                fills[i][j].imageClass = din.readByte();
 if (symCode == 172 || symCode == 173 || symCode == 176 || 
     symCode == 177 || symCode == 178) fills[i][j].imageClass = IMAGE_CLASS6;               
 if (symCode == 179) fills[i][j].imageClass = IMAGE_CLASS7;               
                fills[i][j].numRotations = (byte)(din.readByte()+1);
                //119 thinks it has 4 rots, but it should have 8!
                if (symCode == 119) fills[i][j].numRotations = 8;
                fills[i][j].numFlops = din.readByte();
                int numImages = din.readByte();
                fills[i][j].images = new SymbolImage[numImages];
                int[] imageOffsets = new int[numImages];
                int min = 0;
                for (int k = 0; k < numImages; k++) {
                    imageOffsets[k] = din.readUnsignedShort();
                    if (imageOffsets[k] == 0) {
                        System.err.println("Missing image at symbol "+symCode+", variation "+i+
                                           ", fill "+j+", image "+k);
                    } else if (imageOffsets[k] < min) {
                        throw new IOException("List of image offsets out of order. Expected something "+
                                              "greater than "+(min-1)+" and got "+imageOffsets[k]+
                                              " on symbol "+symCode+", variation "+i+
                                              ", fill "+j+", image "+k);
                    }
                    min = imageOffsets[k] + 6;
                }

                for (int k = 0; k < numImages; k++) {
                    if (imageOffsets[k] == 0) {
                        continue;
                    }
                    if (getCurrentOffset(start,din) != imageOffsets[k]) {
                        throw new IOException("Error reading symbol "+symCode+", variation "+i+
                                              ", fill "+j+", image "+k+" at offset "+
                                              getCurrentOffset(start,din)+
                                              "; was supposed to be at "+imageOffsets[k]);
                    }
                    fills[i][j].images[k] = new SymbolImage();
                    fills[i][j].images[k].framePtV = din.readByte();
                    fills[i][j].images[k].framePtH = din.readByte();
                    fills[i][j].images[k].centerPtV = din.readByte();
                    fills[i][j].images[k].centerPtH = din.readByte();

                    int numBitmapBytes = din.readShort();

                    fills[i][j].images[k].bitmap = new byte[numBitmapBytes];
                    din.readFully(fills[i][j].images[k].bitmap);
                    totalImageCount++;
                }
            }
        }
        if (totalImageCount > 0) {
            setSymbol(symCode,this);
        }
    }

    public int getType() {
    	return type;
    }

    public int getMaxVar() {
    	return fills.length - 1;
    }

    public int getFillVars(int symVar) {
        return fills[symVar].length;
    }

    public int getImageClass(int symVar, int fillVar) {
        return fills[symVar][fillVar].imageClass;
    }

    public int getNumRotations(int symVar, int fillVar) {
        return fills[symVar][fillVar].numRotations;
    }

    public int getNumFlops(int symVar, int fillVar) {
        return fills[symVar][fillVar].numFlops;
    }

    public int getNumVariations() {
        return fills.length;
    }

    public int getNumFills(int symVar) {
        return fills[symVar].length;
    }

    private static int getTranslateImage(int imgClass, int rot, boolean flop) {
    	switch (imgClass) {

    		case IMAGE_CLASS1:
    			if (flop) {
    				return mapClass1Flop[rot].image;
    			} else {
    				return mapClass1[rot].image;
    			}

    		case IMAGE_CLASS2:
    			if (flop) {
    				return mapClass2Flop[rot].image;
    			} else {
    				return mapClass2[rot].image;
    			}

    		case IMAGE_CLASS3A:
    			if (flop) {
    				return mapClass3AFlop[rot].image;
    			} else {
    				return mapClass3A[rot].image;
    			}

    		case IMAGE_CLASS3B:
    			if (flop) {
    				return mapClass3BFlop[rot].image;
    			} else {
    				return mapClass3B[rot].image;
    			}

    		case IMAGE_CLASS4:
    			if (flop) {
    				return mapClass4Flop[rot].image;
    			} else {
    				return mapClass4[rot].image;
    			}

    		case IMAGE_CLASS5:
    			if (flop) {
    				return mapClass5Flop[rot].image;
    			} else {
    				return mapClass5[rot].image;
    			}

    		case IMAGE_CLASS6:
    			if (flop) {
    				return mapClass6Flop[rot].image;
    			} else {
    				return mapClass6[rot].image;
    			}
 
    		case IMAGE_CLASS7:
    			if (flop) {
    				return mapClass7Flop[rot].image;
    			} else {
    				return mapClass7[rot].image;
    			}
 
    		default:
    			return 0;
    	}
    }

    private static int getTranslateTransform(int imgClass, int rot, boolean flop) {
    	switch (imgClass) {

    		case IMAGE_CLASS1:
    			if (flop) {
    				return mapClass1Flop[rot].transform;
    			} else {
    				return mapClass1[rot].transform;
    			}

    		case IMAGE_CLASS2:
    			if (flop) {
    				return mapClass2Flop[rot].transform;
    			} else {
    				return mapClass2[rot].transform;
    			}

    		case IMAGE_CLASS3A:
    			if (flop) {
    				return mapClass3AFlop[rot].transform;
    			} else {
    				return mapClass3A[rot].transform;
    			}

    		case IMAGE_CLASS3B:
    			if (flop) {
    				return mapClass3BFlop[rot].transform;
    			} else {
    				return mapClass3B[rot].transform;
    			}

    		case IMAGE_CLASS4:
    			if (flop) {
    				return mapClass4Flop[rot].transform;
    			} else {
    				return mapClass4[rot].transform;
    			}

    		case IMAGE_CLASS5:
    			if (flop) {
    				return mapClass5Flop[rot].transform;
    			} else {
    				return mapClass5[rot].transform;
    			}

    		case IMAGE_CLASS6:
    			if (flop) {
    				return mapClass6Flop[rot].transform;
    			} else {
    				return mapClass6[rot].transform;
    			}
 
    		case IMAGE_CLASS7:
    			if (flop) {
    				return mapClass7Flop[rot].transform;
    			} else {
    				return mapClass7[rot].transform;
    			}
 
    		default:
    			return NO_TRANSFORM;
    	}
    }

    public SymbolFill getSymbolFill(int symVar, int fillVar) {
        return fills[symVar][fillVar];
    }

    public ImageCoords getCoords (int symVar, int fillVar, int rot, boolean flop, boolean thumbNail) {
if (symVar >= fills.length || fillVar >= fills[symVar].length) return null;
    	SymbolFill fill = fills[symVar][fillVar];
    	int imgClass = fill.imageClass;
    	int image = getTranslateImage(imgClass, rot, flop);
    	int transform = getTranslateTransform(imgClass, rot, flop);

    	SymbolImage im = fill.images[image];
    															// get image coords
    	ImageCoords rtn = new ImageCoords();

    	rtn.framePtH = im.framePtH;
    	rtn.framePtV = im.framePtV;
    	rtn.centerPtH = im.centerPtH;
    	rtn.centerPtV = im.centerPtV;
    	                                                        // adjust coords
    	if (transform == X_TRANSFORM || transform == XY_TRANSFORM) {
    	    rtn.centerPtH = (byte)(rtn.framePtH - rtn.centerPtH);
    	}

    	if (transform == Y_TRANSFORM || transform == XY_TRANSFORM) {
    	    rtn.centerPtV = (byte)(rtn.framePtV - rtn.centerPtV);
        }
    															// scale coords
    	if (thumbNail) {
    		rtn.framePtH /= THUMB_SCALE;
    		rtn.centerPtH /= THUMB_SCALE;
    		rtn.centerPtH /= THUMB_SCALE;
    		rtn.centerPtV /= THUMB_SCALE;
    	}

    	return rtn;
    }

    private void drawPixel (GadgetGraphics g, int xOffset, int yOffset,
                            int bytesPerRow, int byteIndex, byte pixel,
                            boolean thumbNail, Bitmap bitmap,
                            int xbmoffset, int ybmoffset, int[] pixels, PaintInfo paintInfo) {
        int byteX = (bytesPerRow-byteIndex) % bytesPerRow+bytesPerRow;
        int byteY = byteIndex / bytesPerRow;
        if (byteY==0 && byteX != bytesPerRow) byteX -= bytesPerRow;
        for (int i = 0; i < 8; i++) {
            if (((pixel >>> i) & 1) == 1) {
                int x = ((byteX-1)*8+i);
                int y = byteY;
                if (x >= 0 && y >= 0) {
                    if (thumbNail) {
                        if (x%2==0 && y%2==0) {
                            x /= 2;
                            y /= 2;
                            if (pixels != null) {
                                int pixelIndex = ((y+ybmoffset)*paintInfo.width)+(x+xbmoffset);
//if (pixelIndex>0)pixelIndex--;
                                if (pixelIndex >= pixels.length) {
                                    System.err.println("ERROR pixelIndex="+pixelIndex+" len="+pixels.length+" x="+(x+xbmoffset)+" y="+(y+ybmoffset));
                                } else {
                                    pixels[pixelIndex] = paintInfo.color;
                                }
                            } else if (bitmap != null) {
                                bitmap.setBit(x+xbmoffset, y+ybmoffset);
//                                bitmap.getBit(bitmap.getWidth()-1,bitmap.getHeight()-1);
                            } else {
                                x += xOffset;
                                y += yOffset;
                                g.drawLine(x, y, x, y);
                            }
                        }
                    } else {
                        if (pixels != null) {
                            int pixelIndex = ((y+ybmoffset)*paintInfo.width)+(x+xbmoffset);
                            if (pixelIndex >= pixels.length) {
                                System.err.println("ERROR pixelIndex="+pixelIndex+" length="+pixels.length);
                            } else {
                                pixels[pixelIndex] = paintInfo.color;
                            }
                        } else if (bitmap != null) {
                            bitmap.setBit(x+xbmoffset, y+ybmoffset);
//                            bitmap.getBit(bitmap.getWidth()-1,bitmap.getHeight()-1);
                        } else {
                            x += xOffset;
                            y += yOffset;
                            g.drawLine(x, y, x, y);
                        }
                    }
                }
            }
        }
    }

    //This method never called...(?)
    /*public void paint(GadgetGraphics g, int xOffset, int yOffset,
                      int symVar, int fillVar, int rot, boolean flop,
    			      boolean thumbNail) {
        paint(g, xOffset, yOffset, symVar, fillVar, rot, flop,
              thumbNail, null, 0, 0, null, null);
    }*/

    public void paint(GadgetGraphics g, int xOffset, int yOffset,
          int symVar, int fillVar, int rot, boolean flop,
	      boolean thumbNail, Bitmap bitmap,
	      int xbmoffset, int ybmoffset, int[] pixels, PaintInfo paintInfo) {

    	SymbolFill fill = fills[symVar][fillVar];
    	int imgClass = fill.imageClass;
    															// get image
    	int image = getTranslateImage(imgClass, rot, flop);
    	int transform = getTranslateTransform(imgClass, rot, flop);

        if (transform == X_TRANSFORM) {
                transform = NO_TRANSFORM;
        } else if (transform == NO_TRANSFORM) {
                transform = X_TRANSFORM;
        } else if (transform == XY_TRANSFORM) {
                transform = Y_TRANSFORM;
        } else if (transform == Y_TRANSFORM) {
                transform = XY_TRANSFORM;
        }
    	SymbolImage im = fill.images[image];

    	// init draw bitmap
    	int bytesPerRow = 2 * ((im.framePtH + 15) / 16);
    	if (transform != X_TRANSFORM && transform != XY_TRANSFORM) {
    		xOffset -= (((bytesPerRow * 8) - im.framePtH) /*+ (thumbNail ? 1 : 0)*/)/ (thumbNail ? 2 : 1);
    		xbmoffset -= (((bytesPerRow * 8) - im.framePtH) /*+ (thumbNail ? 1 : 0)*/)/ (thumbNail ? 2 : 1);
    	}

    	int count = im.bitmap.length;
    	switch (transform) {

    		case NO_TRANSFORM:												//  no transform
    			for (int i = 0; i < count; i++) {
    				drawPixel(g, xOffset, yOffset, bytesPerRow, i, im.bitmap[i], 
    				    thumbNail, bitmap, xbmoffset, ybmoffset, pixels,
    				    paintInfo);
    			}
    			break;


    		case X_TRANSFORM:				  								//  invert horizontally
    			for (int j = 0; j < count; j += bytesPerRow) {
    				for (int i = 0; i < bytesPerRow; i++) {
    				    int pixel = im.bitmap[bytesPerRow-1-i+j];
    				    if (pixel < 0) {
    				        pixel |= 0x80;
        				    pixel &= 0xff;
        				}
        				drawPixel(g, xOffset, yOffset, bytesPerRow, i+j,
    				        flipByte[pixel], thumbNail,
    				        bitmap, xbmoffset, ybmoffset, pixels, paintInfo);
    				}
    			}
    			break;


    		case Y_TRANSFORM:			  									//  invert vertically
    			for (int j = 0; j < count; j += bytesPerRow) {
    				for (int i = 0; i < bytesPerRow; i++) {
        				drawPixel(g, xOffset, yOffset, bytesPerRow, i+j,
				            im.bitmap[i+count-bytesPerRow-j], thumbNail,
				            bitmap, xbmoffset, ybmoffset, pixels, paintInfo);
    				}
    			}
    			break;


    		case XY_TRANSFORM:			  								//  invert horiz. and vert.
    			for (int i = 0; i < count; i++) {
    			    int pixel = im.bitmap[count-1-i];
				    if (pixel < 0) {
				        pixel |= 0x80;
    				    pixel &= 0xff;
    				}
    				drawPixel(g, xOffset, yOffset, bytesPerRow, i,
				        flipByte[pixel], thumbNail,
				        bitmap, xbmoffset, ybmoffset, pixels, paintInfo);
    			}
    			break;
    	}
    }

}
