package org.signwriting;

import com.dtai.gwt.ButtonGadget;
import com.dtai.gwt.LabelGadget;
import com.dtai.gwt.GadgetDialog;
import com.dtai.gwt.PanelGadget;
import com.dtai.gwt.CheckboxGadget;
import com.dtai.gwt.CheckboxGadgetGroup;
import com.dtai.gwt.ComboBoxGadget;
import com.dtai.gwt.GadgetFlowLayout;
import com.dtai.gwt.GadgetGridLayout;
import com.dtai.gui.StdDialog;
import com.dtai.gwt.BorderGadget;
import java11.awt.event.ActionEvent;
import java11.awt.event.ActionListener;
import java.util.MissingResourceException;

public class ViewPrefsDialog extends StdDialog {
    
    private static final String dialogPrefix = "ViewPrefsDialog.";
    private CheckboxGadget signcb;
    private CheckboxGadget wordcb;
    private CheckboxGadget bothcb;
    private CheckboxGadget largeFontCb;
    private CheckboxGadget medFontCb;
    private CheckboxGadget smallFontCb;
    //private CountryComboBox countries;
    private ColorCombo windowColorCombo;
    private ColorCombo paperColorCombo;
    private ColorCombo inkColorCombo;
    private SignWriter signwriter;
    
    public ViewPrefsDialog(SignWriter signwriter) {
        super(signwriter.getFrame(), "View Preferences", false);
        this.signwriter = signwriter;
        new Buttons(this);
        
        PanelGadget panel = getBody();
        panel.setLayout(new GadgetGridLayout(0,1));
        
        BorderGadget menuFontSel = new BorderGadget();
        menuFontSel.setBorderType(BorderGadget.ETCHED_IN);
        menuFontSel.setBorderThickness(2);
        menuFontSel.setLayout(new GadgetGridLayout(0,1));
        panel.add(menuFontSel);
        
        CheckboxGadgetGroup menuFontGroup = new CheckboxGadgetGroup();
        menuFontSel.add(signcb = new CheckboxGadget("Menus in Signs",true,menuFontGroup));
        menuFontSel.add(wordcb = new CheckboxGadget("Menus in Words",false,menuFontGroup));
        menuFontSel.add(bothcb = new CheckboxGadget("Menus in Signs and Words",false,menuFontGroup));
        if (signwriter.getLanguage() == SignWriter.WORDS) {
            wordcb.setState(true);
        } else if (signwriter.getLanguage() == SignWriter.BOTH) {
            bothcb.setState(true);
        }
        
        BorderGadget displayFontSel = new BorderGadget();
        displayFontSel.setBorderType(BorderGadget.ETCHED_IN);
        displayFontSel.setBorderThickness(2);
        displayFontSel.setLayout(new GadgetGridLayout(0,1));
        panel.add(displayFontSel);
        
        CheckboxGadgetGroup displayFontGroup = new CheckboxGadgetGroup();
        displayFontSel.add(largeFontCb = new CheckboxGadget("Large Font",true,displayFontGroup));
        displayFontSel.add(medFontCb = new CheckboxGadget("Medium Font",false,displayFontGroup));
        displayFontSel.add(smallFontCb = new CheckboxGadget("Small Font",false,displayFontGroup));
        //if (signwriter.getEditFont() == SignFont.small) {
            largeFontCb.setState(true);
        //}
        //THERE IS No medium font yet, but there is a disabled button to show
        //users that one is coming in the dim distant future...
        medFontCb.setEnabled(false);
        //SMALL Fonts are currently disabled too, although they are closer to
        //becoming functional than medium fonts, which are totally nonexistant.
        smallFontCb.setEnabled(false);
        
        BorderGadget typeDirectionSel = new BorderGadget();
        typeDirectionSel.setBorderType(BorderGadget.ETCHED_IN);
        typeDirectionSel.setBorderThickness(2);
        typeDirectionSel.setLayout(new GadgetGridLayout(0,1));
        panel.add(typeDirectionSel);
        
        CheckboxGadgetGroup typeDirectionFontGroup = new CheckboxGadgetGroup();
        CheckboxGadget typeDownCb;
        CheckboxGadget typeAcrossCb;
        typeDirectionSel.add(typeDownCb = new CheckboxGadget("Type Down",true,typeDirectionFontGroup));
        typeDirectionSel.add(typeAcrossCb = new CheckboxGadget("Type Left-to-Right",false,typeDirectionFontGroup));
        typeAcrossCb.setState(true);
        typeDownCb.setEnabled(false);
        
        //THE Panel panel should probably be in GadgetGridBagLayout, so this
        //country selector doesn't have to be the same size as the others.
        //then we can switch the countrySel layout back to GadgetGrid.
        /*BorderGadget countrySel = new BorderGadget();
        countrySel.setBorderType(BorderGadget.ETCHED_IN);
        countrySel.setBorderThickness(2);
        countrySel.setLayout(new GadgetFlowLayout());
        //countrySel.setLayout(new GadgetGridLayout(0,2));
        panel.add(countrySel);
        
        countrySel.add(new LabelGadget("Country:"));
        countrySel.add(countries = new CountryComboBox());*/
        
        BorderGadget colorSel = new BorderGadget();
        colorSel.setBorderType(BorderGadget.ETCHED_IN);
        colorSel.setBorderThickness(2);
        colorSel.setLayout(new GadgetGridLayout(0,2));
        panel.add(colorSel);
        
        colorSel.add(new LabelGadget("Window Color:"));
        colorSel.add(windowColorCombo = new ColorCombo(signwriter.getWindowColor()));
        colorSel.add(new LabelGadget("Paper Color:"));
        colorSel.add(paperColorCombo = new ColorCombo(signwriter.getPaperColor()));
        colorSel.add(new LabelGadget("Ink Color:"));
        colorSel.add(inkColorCombo = new ColorCombo(signwriter.getInkColor()));
        if (signwriter.getEditFont() == SignFont.small) {
            smallFontCb.setState(true);
        }
    }
    
    private void apply() {
        
        signwriter.setWindowColor(windowColorCombo.getColor());
        signwriter.setPaperColor(paperColorCombo.getColor());
        signwriter.setInkColor(inkColorCombo.getColor());
        
        if (signcb.getState()) {
            signwriter.setLanguage(SignWriter.SIGN_LANGUAGE);
        } else if (wordcb.getState()) {
            signwriter.setLanguage(SignWriter.WORDS);
        } else {
            signwriter.setLanguage(SignWriter.BOTH);
        }
        //WHEN The country combobox gets used in the future:
        //do something like: signwriter.setCountry(countries.getSelectedIndex());
        //which sets which dictionary will be used, and which keyboard, etc...
        //For now, the countries combobox always has the USA selected!
        if (largeFontCb.getState()) {
            signwriter.setEditFont(SignFont.large);
        } else {
            signwriter.setEditFont(SignFont.small);
        }
    }

    class Buttons { // Symantec Java compiler bug forced me to
                    // create this bogus class "Buttons"
        
        ViewPrefsDialog dialog;
        
        public Buttons(final ViewPrefsDialog dialog) {
            this.dialog = dialog;
            
            addButton("OK",
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        dialog.apply();
                        dialog.hide();
                    }
                }
            );

            addButton("Apply",
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        dialog.apply();
                    }
                }
            );
            
            addButton("Cancel",
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        dialog.hide();
                    }
                }
            );
        }
        
        private void addButton(String label, ActionListener listener) {
            ButtonGadget button = new ButtonGadget();
            String resourceKey = dialogPrefix+"Buttons."+Resources.labelToKey(label);
            Resources.apply(resourceKey, label, button);
            dialog.addButtonGadget(button);
            button.addActionListener(listener);
        }
    }
}
